#  -*- coding: utf-8 -*-


import os
import sys
import pwd
import grp
import difflib
import stat
import errno
import pkg_resources
from ctypes import CDLL
from datetime import datetime
from collections import OrderedDict
from shutil import copytree, copy2, copystat, move, rmtree
from signal import SIGCONT, SIGKILL, SIGSTOP
from subprocess import check_output, Popen, PIPE, STDOUT
from multiprocessing import Process, Queue, Pipe
from os.path import basename, dirname, exists, expanduser, expandvars, getsize, \
                    isabs, isdir, isfile, islink, join, normpath, relpath
from string import whitespace, punctuation
from time import asctime, localtime, sleep, strftime, time, tzname

from common import *


########################################################################
##### Module variables
use_wide_chars = False


########################################################################
###### LFM
def get_lfm_data_file_contents(filename):
    return str(pkg_resources.resource_string('lfm', 'etc/' + filename), 'UTF-8')


########################################################################
##### Decorators
# Decorator for public callable actions
from functools import wraps
from inspect import getmembers, isfunction

def public(f):
    @wraps(f)
    def wrapper(*a, **k):
        return f(*a, **k)
    setattr(wrapper, 'public', True)
    return wrapper

def get_public_actions():
    import actions
    fns = getmembers(actions, isfunction)
    return [name for name, fn in fns if hasattr(fn, 'public')]

def is_public_api(fn):
    return hasattr(fn, 'public')

# Decorator for catching OSError
def catch_os_exception(f):
    @wraps(f)
    def wrapper(*a, **k):
        try:
            return f(*a, **k)
        except OSError as err:
            return err
    return wrapper


########################################################################
##### ConfigParser with Comments and Header
from configparser import ConfigParser

class ConfigParserWithComments(ConfigParser):
    def __init__(self, header):
        self._header = header
        self.optionxform = str
        super(ConfigParser, self).__init__()

    def write(self, fp):
        if self._header:
            fp.write('{}\n\n'.format(self._header))
        if self._defaults:
            fp.write('[{}]\n'.format(ConfigParser.DEFAULTSECT))
            for (key, value) in self._defaults.items():
                self._write_item(fp, key, value)
            fp.write('\n')
        for section in self._sections:
            fp.write('[{}]\n'.format(section))
            for (key, value) in self._sections[section].items():
                self._write_item(fp, key, value)
            fp.write('\n')

    def _write_item(self, fp, key, value):
        if key == '#' and value:
            for l in value.split('\n'):
                fp.write('# {}\n'.format(l))
        else:
            fp.write('{}: {}\n'.format(key, str(value).replace('\n', '\n\t')))


########################################################################
##### PathContents
class PathContents:
    def __init__(self, fs, basepath):
        self.basepath = basepath
        self.__entries = dict()
        for f in fs:
            try:
                if islink(f):
                    self.__entries[f] = (f, 0, '')
                elif isdir(f):
                    self.__entries[f] = (f, getsize(f), '')
                    self.__fill_contents(f)
                else:
                    self.__entries[f] = (f, getsize(f), '')
            except OSError as err:
                self.__entries[f] = (f, 0, self.__format_err(f, err))
        self.__entries = sorted(self.__entries.values())
        self.length = len(fs)
        self.tlength = len(self.__entries)
        self.tsize = sum([f[1] for f in self.__entries]) or 1

    def __fill_contents(self, path):
        for root, dirs, files in os.walk(path, topdown=False, onerror=self.__on_error):
            for f in dirs + files:
                fullpath = join(root, f)
                try:
                    if islink(fullpath):
                        self.__entries[fullpath] = (fullpath, 0, '')
                    else:
                        self.__entries[fullpath] = (fullpath, getsize(fullpath), '')
                except OSError as err:
                    self.__entries[fullpath] = (fullpath, 0, self.__format_err(fullpath, err))

    def __on_error(self, err):
        fullpath = join(self.basepath, err.filename)
        self.__entries[fullpath] = (fullpath, 0, self.__format_err(fullpath, err))

    def __format_err(self, filename, err):
        return Exception('[Errno {}] {}: \'{}\''.format(err.errno, err.strerror, filename.replace(self.basepath, '')))

    def __repr__(self):
        return 'PathContents[Base:"{}" with {} entries (Total: {} items, {:.2f} KB)]' \
            .format(self.basepath, self.length, self.tlength, self.tsize/1024)

    def remove_files(self, fs):
        new, size = list(), 0
        for f, s, e in self.__entries:
            if f not in fs:
                new.append((f, s, e))
                size += s
        self.__entries = new
        self.length = self.length # not really, but not important either
        self.tlength = len(self.__entries)
        self.tsize = size

    @property
    def entries(self):
        return sorted(self.__entries, reverse=False)

    @property
    def entries_rev(self):
        return sorted(self.__entries, reverse=True)


########################################################################
##### DirsTree
class DirsTree:
    def __init__(self, path, dotfiles):
        self._dotfiles = dotfiles
        self.build(path)

    def __get_graph(self, path):
        """return a OrderedDict with tree structure"""
        d, expanded = dict(), None
        while path:
            if path==os.sep and os.sep in d:
                break
            d[path] = (get_dirs(path, self._dotfiles), expanded)
            expanded = basename(path)
            path = dirname(path)
        return OrderedDict(sorted(d.items(), key=lambda t: t[0]))

    def __get_node(self, i, td, parent):
        """expand branch. Each node has (name, depth, fullname))"""
        dirs, expanded_node = td[list(td.keys())[i]]
        if not expanded_node:
            return list()
        lst = list()
        for d in dirs:
            lst.append((d, i, join(parent, d)))
            if d == expanded_node:
                lst2 = self.__get_node(i+1, td, join(parent, d))
                if lst2 is not None:
                    lst.extend(lst2)
        return lst

    def __getitem__(self, i):
        return self._data[i]

    def __len__(self):
        return len(self._data)

    def build(self, path):
        """build list with tree structure"""
        td = self.__get_graph(path)
        self._data = [(os.sep, -1, os.sep)]
        self._data.extend(self.__get_node(0, td, os.sep))
        self.path = path

    def regenerate_from_pos(self, newpos):
        """regenerate tree when changing to a new directory and return it"""
        newpath = self._data[newpos][2]
        self.build(newpath)
        return newpath

    @property
    def pos(self):
        """return position of current dir"""
        for i in range(len(self._data)):
            if self.path == self._data[i][2]:
                return i
        else:
            return -1

    @property
    def cur_depth(self):
        return self.get_depth(self.pos)

    def get_depth(self, pos):
        return self._data[pos][1]

    @property
    def is_first_sibling(self):
        return self.cur_depth != self.get_depth(self.pos-1)

    @property
    def first_sibling_pos(self):
        newpos = self.pos
        while True:
            if newpos-1 < 0 or self.cur_depth != self.get_depth(newpos-1):
                break
            newpos -= 1
        return newpos

    @property
    def is_last_sibling(self):
        return self.cur_depth != self.get_depth(self.pos+1)

    @property
    def last_sibling_pos(self):
        newpos = self.pos
        while True:
            if newpos+1 == len(self) or self.cur_depth != self.get_depth(newpos+1):
                break
            newpos += 1
        return newpos

    @property
    def parent_pos(self):
        for i in range(self.pos-1, -1, -1):
            if self.get_depth(i) == self.cur_depth-1:
                break
        return i

    @property
    def has_children_dirs(self):
        return len(get_dirs(self.path, self._dotfiles)) > 0

    def to_child(self):
        newpath = None
        child_dirs = get_dirs(self.path, self._dotfiles)
        if len(child_dirs) > 0:
            newpath = join(self.path, child_dirs[0])
            self.build(newpath)
        return newpath

    def show_tree(self, a=0, z=-1):
        """show an ascii representation of the tree. Not used in lfm"""
        if z>len(self._data) or z==-1:
            z = len(self._data)
        for i in range(a, z):
            name, depth, fullname = self._data[i]
            if fullname == self.path:
                name += ' <====='
            if name == os.sep:
                print(' ' + name)
            else:
                print(' | ' * depth + ' +- ' + name)


########################################################################
##### Paths
def get_realpath(tab):
    if tab.fs[tab.i].is_link:
        try:
            return '-> ' + os.readlink(join(tab.dirname, tab.current_filename))
        except OSError:
            return tab.fs.path_str + '/' + tab.current_filename
    else:
        return tab.fs.path_str + '/' + tab.current_filename

def get_relpath(path, base):
    return relpath(path, base) if path.startswith(base) else path

def get_norm_path(path, basedir):
    path = expandvars(expanduser(path))
    path = path if isabs(path) else join(basedir, path)
    return normpath(path)

def get_dir_size(dirname):
    try:
        buf = check_output(['du', '-sb', dirname], stderr=STDOUT, universal_newlines=True)
        return int(buf.split()[0])
    except:
        return -1

def get_dirs(path, dotfiles):
    try:
        if dotfiles:
            ds = [d for d in os.listdir(path) if isdir(join(path, d))]
        else:
            ds = [d for d in os.listdir(path) if d[0]!='.' and isdir(join(path, d))]
    except OSError:
        return list()
    return sorted(ds)


########################################################################
##### Files
def get_filetype(pfile):
    """Returns the type of a file as FileType"""
    lmode = os.lstat(pfile).st_mode
    if stat.S_ISDIR(lmode):
        return FileType.dir
    if stat.S_ISLNK(lmode):
        try:
            mode = os.stat(pfile)[stat.ST_MODE]
        except OSError:
            return FileType.nlink
        else:
            return FileType.link2dir if stat.S_ISDIR(mode) else FileType.link
    if stat.S_ISCHR(lmode):
        return FileType.cdev
    if stat.S_ISBLK(lmode):
        return FileType.bdev
    if stat.S_ISFIFO(lmode):
        return FileType.fifo
    if stat.S_ISSOCK(lmode):
        return FileType.socket
    if stat.S_ISREG(lmode) and (lmode & 0o111):
        return FileType.exe
    else:
        return FileType.reg # if no other type, regular file

def get_file_info(filename):
    try:
        return check_output(['file', '-b', filename], stderr=STDOUT, universal_newlines=True).strip()
    except:
        return 'no type information'

def backup_file(src, backup_ext):
    dest = src + backup_ext
    if exists(dest):
        raise FileExistsError('Cannot backup file:\n"{}" already exists'.format(dest))
    try:
        return copy_bulk(src, dest)
    except OSError as err:
        raise Exception('Cannot backup file:\n{}'.format(err))

def get_file_diff(file_old, file_new, diff_type):
    try:
        d0 = datetime.fromtimestamp(os.stat(file_old).st_mtime)
        date_old = d0.strftime('    %Y-%m-%d %H:%M:%S.%f ') + tzname[0]
        d1 = datetime.fromtimestamp(os.stat(file_new).st_mtime)
        date_new = d1.strftime('    %Y-%m-%d %H:%M:%S.%f ') + tzname[0]
        # with open(file_old).readlines() as buf0, open(file_new).readlines() as buf1:
        with open(file_old) as f0, open(file_new) as f1:
            buf0, buf1 = f0.readlines(), f1.readlines()
            if diff_type == 'context':
                diff = difflib.context_diff(buf0, buf1, file_old, file_new, date_old, date_new)
            elif diff_type == 'unified':
                diff = difflib.unified_diff(buf0, buf1, file_old, file_new, date_old, date_new)
            elif diff_type == 'ndiff':
                diff = difflib.ndiff(buf0, buf1)
            return ''.join(diff)
    except:
        return ''

def copy_file(filename, basepath, destdir, overwrite=False):
    partial = filename.replace(basepath, '')
    dest = join(destdir, partial) if isdir(destdir) else destdir
    try:
        dest_exists = exists(dest)
        if dest_exists and not overwrite:
            raise LFMFileExistsError(dest)
        if islink(filename) or isfile(filename):
            if dest_exists:
                os.unlink(dest)
            copy2(filename, dest, follow_symlinks=False)
        elif isdir(filename):
            if dest_exists:
                try:
                    os.rmdir(dest)
                except OSError as err:
                    if err.errno != errno.ENOTEMPTY: # Directory not empty
                        raise
            try:
                os.mkdir(dest)
            except OSError as err:
                if err.errno != errno.EEXIST: # File exists
                    raise
            try:
                st = os.lstat(src)
                os.chown(dest, st[stat.ST_UID], st[stat.ST_GID])
                copystat(src, dest, follow_symlinks=False)
            except:
                pass
        else:
            raise Exception('Cannot copy file!\nCannot copy special file: \'{}\''.format(partial))
    except OSError as err:
        raise Exception('Cannot copy file!\n[Errno {}] {}: \'{}\''.format(err.errno, err.strerror, partial))

def move_file(filename, basepath, destdir, overwrite=False):
    # alternative method instead of copy & delete. Faster but less control
    partial = filename.replace(basepath, '')
    dest = join(destdir, partial) if isdir(destdir) else destdir
    try:
        if exists(dest) and not overwrite:
            raise LFMFileExistsError(dest)
        move(filename, dest)
    except OSError as err:
        raise Exception('Cannot move file!\n[Errno {}] {}: \'{}\''.format(err.errno, err.strerror, partial))

def delete_file(filename, basepath):
    try:
        if islink(filename):
            os.unlink(filename)
        elif isdir(filename):
            os.rmdir(filename)
        else:
            os.unlink(filename)
    except OSError as err:
        raise Exception('Cannot delete file!\n[Errno {}] {}: \'{}\''.format(err.errno, err.strerror, filename.replace(basepath, '')))

@catch_os_exception
def make_dir(newdir):
    os.makedirs(newdir)

@catch_os_exception
def touch_file(filename):
    with open(filename, 'a'):
        os.utime(filename)

@catch_os_exception
def rename_file(src, dest):
    os.rename(src, dest)

@catch_os_exception
def link_create(linkname, pointto):
    os.symlink(pointto, linkname)

@catch_os_exception
def link_edit(linkname, pointto):
    os.unlink(linkname)
    os.symlink(pointto, linkname)


##### copy, overwrite, delete
def copy_bulk(src, dest):
    if isdir(src):
        copytree(src, dest, symlinks=True)
    elif isfile(src):
        copy2(src, dest)

def overwrite_vfsfile(src, dest):
    copystat(dest, src) # copy attrs
    move(src, dest)

def delete_bulk(path, ignore_errors=False):
    if isdir(path):
        rmtree(path, ignore_errors=ignore_errors)
    elif isfile(path):
        if ignore_errors:
            try:
                os.unlink(path)
            except OSError:
                pass
        else:
            os.unlink(path)


########################################################################
##### Un/compress
from compress import get_compressed_file_engine, packagers_by_type, PackagerTAR

def uncompress_dir(filename, destdir):
    if not isfile(filename):
        raise Exception('It\'s not a file: {}'.format(basename(filename)))
    c = get_compressed_file_engine(filename)
    if c is None:
        raise Exception('Cannot uncompress this file type: {}'.format(basename(filename)))
    res, err = run_in_cli(c.cmd_uncompress, destdir)
    if err:
        raise Exception(err)
    return res

def compress_dir(dir, typ):
    if not isdir(dir):
        raise Exception('It\'s not a directory: {}'.format(basename(dir)))
    c = packagers_by_type[typ](dir)
    if c is None:
        raise Exception('Cannot compress: {}'.format(basename(dir)))
    res, err = run_in_cli(c.cmd_compress, dirname(dir))
    if err:
        raise Exception(err)
    return res

def compress_uncompress_file(filename, typ):
    if not isfile(filename):
        raise Exception('It\'s not a file: {}'.format(basename(filename)))
    c = get_compressed_file_engine(filename)
    if c is None or isinstance(c, PackagerTAR):
        cmd = packagers_by_type[typ](filename).cmd_compress
    elif c.type == typ:
        cmd = c.cmd_uncompress
    else:
        raise Exception('Cannot un/compress \'{}\' with type {}'.format(basename(filename), typ))
    res, err = run_in_cli(cmd, dirname(filename))
    if err:
        raise Exception(err)
    return res


########################################################################
##### FileSystems
def get_filesystems_info():
    return check_output(['df', '-h'], stderr=STDOUT, universal_newlines=True)

def get_mount_points():
    """return system mount points as list of (mountpoint, device, fstype).
    Compatible with linux and solaris"""
    buf = check_output(['mount'], stderr=STDOUT, universal_newlines=True).strip()
    lst = [(e.split()[2], e.split()[0], e.split()[4]) for e in buf.split('\n')]
    return sorted(lst, reverse=True)

def get_mountpoint_for_file(filename):
    try:
        for m, d, t in get_mount_points():
            if filename.find(m) != -1:
                return (m, d, t)
        else:
            raise
    except:
        return ('/', '<unknown>', '<unknown>')


########################################################################
##### Users & Groups
def get_user_fullname(user):
    try:
        return pwd.getpwnam(user)[4]
    except KeyError:
        return '<unknown user name>'

def get_owners():
    """get a list with the users defined in the system"""
    return sorted([e[0] for e in pwd.getpwall()])

def get_groups():
    """get a list with the groups defined in the system"""
    return sorted([e[0] for e in grp.getgrall()])


########################################################################
##### Binary programs
def get_binary_programs():
    return sorted({f for p in os.getenv('PATH').split(':') if isdir(p) for f in os.listdir(p) if isfile(join(p, f))})


########################################################################
##### String formatting
def size2str(size):
    """Converts a file size into a string"""
    if size >= 1000000000:
        return str(size//1048576) + 'M' # 1024*1024
    elif size >= 10000000:
        return str(size//1024) + 'K'
    else:
        return str(size)

def num2str(num):
    # Thanks to "Fatal" in #pys60
    num = str(num)
    return (len(num) < 4) and num or (num2str(num[:-3])+","+num[-3:])

def time2str(t, short=True):
    """Converts a file time into a string"""
    if -15552000 < (time() - t) < 15552000:
        # date < 6 months from now, past or future
        fmt = '%d %b %H:%M' if short else '%a %b %d %H:%M'
    else:
        fmt = '%d %b  %Y' if short else '%a  %d %b %Y'
    return strftime(fmt, localtime(t))

def time2str_full(t):
    return asctime(localtime(t))

def perms2str(perms):
    """Converts a file permisions into a string"""
    return stat.filemode(perms)[1:].lower()

def type2str(ftype):
    """Converts a file type into a string"""
    return FILETYPES[ftype][0]

def rdev2str(rdev):
    """Converts a device file numbers into a string"""
    return ('%d,%d' % rdev).rjust(7)

def owner2str(uid):
    try:
        return pwd.getpwuid(uid).pw_name
    except KeyError:
        return str(uid)

def group2str(gid):
    try:
        return grp.getgrgid(gid).gr_name
    except KeyError:
        return str(gid)

def str2perms(perms):
    ps = 0
    for i, p in enumerate(reversed(perms)):
        if p == 'x':
            ps += 1 * 8**(i//3)
        elif p == 'w':
            ps += 2 * 8**(i//3)
        elif p == 'r':
            ps += 4 * 8**(i//3)
        elif p == 't' and i == 0:
            ps += 1 * 8**3
        elif p == 's':
            if i == 6:
                ps += 4 * 8**3
            elif i == 3:
                ps += 2 * 8**3
    return ps


########################################################################
##### Support for wide chars
def length(text):
    return wchar_len(text) if use_wide_chars else len(text)

def max_length(entries):
    return max(map(wchar_len if use_wide_chars else len, entries))

def text2wrap(*args, **kw):
    return text2wrap_wchar(*args, **kw) if use_wide_chars else text2wrap_normal(*args, **kw)

def text2wrap_normal(text, max_width, ljust=True, start_pct=.66, sep='~', fill=True):
    """Returns a displayable string from an attribute"""
    length = len(text)
    if length <= max_width:
        if ljust:
            return text.ljust(max_width) if fill else text
        else:
            return text.ljust(max_width) if ljust else text.rjust(max_width)
    else:
        until = int(max_width*start_pct)
        return text[:until] + sep + text[-(max_width-until-1):]

def text2wrap_wchar(text, max_width, ljust=True, start_pct=.66, sep='~', fill=True):
    """Returns a displayable string from an attribute"""
    length = wchar_len(text)
    if length <= max_width:
        if ljust:
            return wchar_ljust(text, max_width) if fill else text
        else:
            return wchar_ljust(text, max_width) if ljust else wchar_rjust(text, max_width)
    else:
        return wchar_fill(text, max_width, start_pct, sep)

def wchar_len(text):
    libc = CDLL('libc.so.6')
    # assert libc.wcswidth('世界')==4
    return len(text) if len(text)>127 else libc.wcswidth(text) # BUG: wcswidth doesn't accept string with 128+ chars

def wchar_ljust(text, max_width):
    return text + ' '*(max_width-wchar_len(text))

def wchar_rjust(text, max_width):
    return ' '*(max_width-wchar_len(text)) + text

def wchar_fill(text, max_width, pct, sep):
    pos = int(max_width*pct)
    buf1 = buf2 = ''
    for c in text:
        if wchar_len(buf1) + wchar_len(c) > pos:
            break
        buf1 += c
    buf1 += sep
    len_buf1 = wchar_len(buf1)
    for c in text[::-1]:
        if len_buf1 + wchar_len(c) + wchar_len(buf2) > max_width:
            break
        buf2 += c
    return buf1+buf2[::-1]


########################################################################
##### Navigate through paths
stepchars = whitespace + punctuation

def prev_step(text, pos):
    pos = max(0, pos-2)
    while pos > 0 and text[pos] not in stepchars:
        pos -=1
    return pos+1 if pos > 0 else 0

def next_step(text, pos):
    pos += 1
    l = len(text)
    while pos < l and text[pos] not in stepchars:
        pos +=1
    return pos+1 if pos < l else l


########################################################################
##### Running commands in shell
import curses

def escape_str(buf):
    if buf.find('"') != -1:
        return '\'{}\''.format(buf.replace('"', '\\"'))
    else:
        return '"{}"'.format(buf)

def escape_command(cmd, filename, background):
    filename = filename.replace('$', '\$')
    return '{} {}{}'.format(cmd, escape_str(filename), ' >/dev/null 2>&1 &' if background else '')

def run_on_current_file(program, filename, background=False):
    curses.endwin()
    os.system(escape_command(program, filename, background))
    curses.curs_set(0)

def run_in_background(cmd, path):
    cmd = 'cd {} && {} >/dev/null 2>&1 &'.format(escape_str(path), cmd)
    curses.endwin()
    os.system(cmd)
    curses.curs_set(0)

def run_shell(shell, path):
    curses.endwin()
    os.system('cd "{}" && {}'.format(path.replace('"', '\\"'), shell))
    curses.curs_set(0)

def run_in_cli(cmd, path):
    return Popen(cmd, shell=True, cwd=path, stdout=PIPE, stderr=PIPE, universal_newlines=True).communicate()


######################################################################
##### Process classes
from ui_widgets import CursorAnimation, DialogConfirm, DialogConfirmAll, DialogConfirmAllNone, \
                       DialogError, DialogMessagePanel, DialogProgress1Panel, DialogProgress2Panel

##### ProcessCommand
class ProcessCommand:
    def __init__(self, title, subtitle, cmd, path=None):
        self.title = title
        self.subtitle = subtitle
        self.cmd = cmd
        self.path = path
        self.proc = None
        self.status = None
        self.dialog = DialogMessagePanel(self.title, self.subtitle)
        self.animation = CursorAnimation()

    def check_stop(self):
        if self.dialog.check_key() == 0x03:
            self.proc.send_signal(SIGSTOP)
            self.dialog.hide()
            if DialogConfirm('Stop process', '{} {}'.format(self.title, self.subtitle), 0) == 1:
                self.proc.kill()
                return True
            else:
                self.dialog.show()
                self.proc.send_signal(SIGCONT)
        return False

    def run(self):
        self.dialog.show()
        results, errors = None, None
        self.proc = Popen(self.cmd, shell=True, cwd=self.path, stdout=PIPE, stderr=PIPE, universal_newlines=True)
        if self.proc.poll(): # process can finish too fast
            while True:
                self.animation.next()
                self.status = self.proc.poll()
                if self.status is not None:
                    break
                if self.check_stop():
                    self.status = -100
                    break
                sleep(0.05)
        self.dialog.hide()
        results, errors = self.proc.communicate()
        return self.status, results, errors


##### ProcessFuncLoop
class ProcessFuncLoop:
    def __init__(self, title, fn, lst):
        self.title = title
        self.fn = fn
        self.lst = lst
        self.i, self.n = 0, len(lst)
        self.proc = None
        self.animation = CursorAnimation()

    def check_stop(self):
        if self.dialog.check_key() == 0x03:
            os.kill(self.proc.pid, SIGSTOP)
            self.dialog.hide()
            if DialogConfirm('Stop process', 'Stop "{}"?'.format(self.title), 0) == 1:
                os.kill(self.proc.pid, SIGKILL)
                return True
            else:
                self.dialog.show()
                os.kill(self.proc.pid, SIGCONT)
        return False

    def run_func(self, fn, qin, qout, qerr, conn):
        while not qin.empty():
            if qin.empty():
                break
            try:
                ret = self.do_run(fn, qin, conn)
            except Exception as err:
                qout.put_nowait(None)
                qerr.put_nowait(err)
                if not isinstance(err, LFMFileSkipped):
                    conn.send((ProcCode.error, str(err)))
            else:
                qout.put_nowait(ret)
                qerr.put_nowait(None)
            # sleep(0.001)
        conn.send((ProcCode.end, ))
        sleep(0.25)

    def run(self):
        self.prepare()
        self.dialog.show()
        qin, qout, qerr = Queue(maxsize=self.n), Queue(maxsize=self.n), Queue(maxsize=self.n)
        for a in self.lst:
            qin.put(a)
        conn_parent, conn_child = Pipe()
        self.proc = Process(target=self.run_func, args=(self.fn, qin, qout, qerr, conn_child))
        self.proc.start()
        while True:
            if conn_parent.poll():
                buf = conn_parent.recv()
                if buf[0] == ProcCode.end:
                    conn_parent.send(ProcCodeConfirm.stop)
                    status = ProcCode.end
                    rets, errs = list(), list()
                    while not qout.empty():
                        rets.append(qout.get_nowait())
                    while not qerr.empty():
                        errs.append(qerr.get_nowait())
                    break
                elif buf[0] == ProcCode.error:
                    self.dialog.hide()
                    msg = buf[1] if buf[1].startswith('Cannot') else 'Cannot {}!\n{}'.format(self.title.lower(), buf[1])
                    DialogError(msg)
                    self.dialog.show()
                elif buf[0] == ProcCode.next:
                    self.display_next(*buf[1:])
                    ans = self.confirm_pre()
                    conn_parent.send(ans)
                    if ans == ProcCodeConfirm.stop:
                        status, rets, errs = ProcCode.stopped, None, None
                        break
                elif buf[0] == ProcCode.confirm:
                    ans = self.confirm_post(buf[1])
                    conn_parent.send(ans)
                    if ans == ProcCodeConfirm.stop:
                        status, rets, errs = ProcCode.stopped, None, None
                        break
            if self.check_stop():
                status, rets, errs = ProcCode.stopped, None, None
                break
            self.animation.next()
            # sleep(0.05)
        self.dialog.hide()
        try:
            os.kill(self.proc.pid, SIGKILL)
        except ProcessLookupError:
            pass
        return status, rets, errs

    def prepare(self):
        self.dialog = DialogProgress1Panel(self.title)

    def display_next(self, text, *args):
        self.i += 1
        self.dialog.update(text, self.i, self.n)

    def confirm_pre(self):
        return ProcCodeConfirm.ok

    def confirm_post(self, *args):
        return ProcCodeConfirm.ok

    def do_run(self, fn, qin, conn):
        args = qin.get_nowait()
        conn.send((ProcCode.next, args[0]))
        _ = conn.recv()
        return fn(*args)


##### ProcessFuncDeleteLoop
class ProcessFuncDeleteLoop(ProcessFuncLoop):
    def __init__(self, title, fn, lst, tot_size, confirm):
        super(ProcessFuncDeleteLoop, self).__init__(title, fn, lst)
        self.tot_size = tot_size
        self.confirm = confirm

    def prepare(self):
        self.dialog = DialogProgress2Panel(self.title)
        self.acc_size = 0

    def display_next(self, text, size, *args):
        self.i += 1
        self.acc_size += size
        self.cur_elm = text
        self.dialog.update(text, self.i, self.n, self.acc_size, self.tot_size)

    def confirm_pre(self):
        if self.confirm:
            self.dialog.hide()
            ans = DialogConfirmAll(self.title, 'Delete \'{}\'?'.format(self.cur_elm), default=1)
            if ans == 1:
                ret = ProcCodeConfirm.ok
            elif ans == 2:
                self.confirm = False
                ret = ProcCodeConfirm.ok
            elif ans == 0:
                ret = ProcCodeConfirm.skip
            else: # ans == -1
                ret = ProcCodeConfirm.stop
            self.dialog.show()
            return ret
        else:
            return ProcCodeConfirm.ok

    def do_run(self, fn, qin, conn):
        filename, size, err, basepath = qin.get_nowait()
        conn.send((ProcCode.next, filename.replace(basepath, ''), size))
        ans = conn.recv()
        if err:
            raise err
        if ans == ProcCodeConfirm.ok:
            return fn(filename, basepath)
        elif ans == ProcCodeConfirm.skip:
            raise LFMFileSkipped(filename)
        else:
            return None


##### ProcessFuncCopyLoop
class ProcessFuncCopyLoop(ProcessFuncLoop):
    def __init__(self, title, fn, lst, tot_size, confirm_overwrite):
        super(ProcessFuncCopyLoop, self).__init__(title, fn, lst)
        self.tot_size = tot_size
        self.overwrite = None if confirm_overwrite else ProcCodeConfirm.ok

    def prepare(self):
        self.dialog = DialogProgress2Panel(self.title)
        self.acc_size = 0

    def display_next(self, text, size, *args):
        self.i += 1
        self.acc_size += size
        self.cur_elm = text
        self.dialog.update(text, self.i, self.n, self.acc_size, self.tot_size)

    def confirm_post(self, filename):
        if self.overwrite is None:
            self.dialog.hide()
            ans = DialogConfirmAllNone(self.title, 'Overwrite \'{}\'?'.format(filename), default=1)
            if ans == 1:
                ret = ProcCodeConfirm.ok
            elif ans == 2:
                ret = self.overwrite = ProcCodeConfirm.ok
            elif ans == 0:
                ret = ProcCodeConfirm.skip
            elif ans == -2:
                ret = self.overwrite = ProcCodeConfirm.skip
            else: # ans == -1
                ret = ProcCodeConfirm.stop
            self.dialog.show()
            return ret
        else:
            return self.overwrite

    def do_run(self, fn, qin, conn):
        filename, size, err, basepath, destdir = qin.get_nowait()
        conn.send((ProcCode.next, filename.replace(basepath, ''), size))
        ans = conn.recv()
        if err:
            raise err
        if ans == ProcCodeConfirm.ok:
            try:
                return fn(filename, basepath, destdir)
            except LFMFileExistsError as err:
                conn.send((ProcCode.confirm, err))
                ans = conn.recv()
                if ans == ProcCodeConfirm.ok:
                    return fn(filename, basepath, destdir, overwrite=True)
                elif ans == ProcCodeConfirm.skip:
                    raise LFMFileSkipped(filename)
                else:
                    return None
        else:
            return None


########################################################################



# -*- coding: utf-8 -*-


import os
from glob import glob, escape
from operator import attrgetter
from tempfile import mkdtemp, mkstemp
from os.path import abspath, basename, dirname, exists, expanduser, isabs, isdir, join, normpath, splitext

from compress import check_compressed_vfs, get_compressed_file_engine
from utils import copy_bulk, delete_bulk, get_filetype, overwrite_vfsfile, \
                  ProcessCommand, run_in_cli, text2wrap, size2str, time2str, \
                  perms2str, owner2str, group2str, type2str, rdev2str
from common import *


########################################################################
##### FS Configuration
class FSConfig:
    def __init__(self):
        self.filters = ''
        self.show_dotfiles = True
        self.sort_type = SortType.byName
        self.sort_reverse = False
        self.sort_mix_dirs = False
        self.sort_mix_cases = True

    def fill_with_app(self, cfg):
        self.show_dotfiles = cfg.options.show_dotfiles
        self.sort_type = cfg.options.sort_type
        self.sort_reverse = cfg.options.sort_reverse
        self.sort_mix_dirs = cfg.options.sort_mix_dirs
        self.sort_mix_cases = cfg.options.sort_mix_cases

    def get_sort_key(self):
        if self.sort_type == SortType.none:
            return lambda x: 0
        elif self.sort_type == SortType.byName:
            return lambda x: (attrgetter('name')(x).lower() if self.sort_mix_cases else attrgetter('name')(x))
        elif self.sort_type == SortType.byExt:
            return lambda x: (attrgetter('ext')(x).lower() if self.sort_mix_cases else attrgetter('ext')(x))
        elif self.sort_type == SortType.byPath:
            return lambda x: (attrgetter('path_str')(x).lower() if self.sort_mix_cases else attrgetter('path_str')(x))
        elif self.sort_type == SortType.bySize:
            return attrgetter('size')
        elif self.sort_type == SortType.byMTime:
            return attrgetter('mtime')
        else:
            raise KeyError


########################################################################
##### FSEntry
class FSEntry:
    """File System entry"""

    def __init__(self, pfile):
        self.pfile = pfile
        self.pdir = dirname(pfile)
        self.name = basename(pfile)
        self.name_noext, self.ext = splitext(self.name)
        if self.name_noext.endswith('.tar'):
            self.name_noext = self.name_noext.replace('.tar', '')
            # self.ext = '.tar' + self.ext
        st = os.lstat(pfile)
        self.size = st.st_size
        self.mode = st.st_mode
        self.owner = st.st_uid
        self.group = st.st_gid
        self.mtime = st.st_mtime
        self.stat = st
        self.type = get_filetype(pfile)
        if self.type in (FileType.cdev, FileType.bdev):
            try:
                r = st.st_rdev
                self.rdev = r >> 8, r & 255
            except AttributeError:
                self.rdev = 0, 0
        else:
            self.rdev = 0, 0
        # string representation of attributes
        self.path_str = str(pfile)
        self.size_str = size2str(self.size)
        self.mtime_str = time2str(self.mtime)
        self.mtime2_str = time2str(self.mtime, short=False)
        self.mode_str = perms2str(self.mode)
        self.owner_str = owner2str(self.owner)
        self.group_str = group2str(self.group)
        self.type_str = type2str(self.type)
        self.rdev_str = rdev2str(self.rdev)

    def format(self, fields, sep='|'):
        # fields is a list of tuples [('attribute', length)]
        #   ls = ['{0.%s:%ds}' % (attr, length) for attr, length in fields]
        #   fmt = sep.join(ls)
        #   return fmt.format(self)
        ls = []
        for attr, length in fields:
            ljust = True
            if attr in ('size', ):
                if self.type in (FileType.cdev, FileType.bdev):
                    attr, length = 'rdev_str', 7
                else:
                    attr, ljust = 'size_str', False
            elif attr in ('mtime', 'mtime2', 'mode', 'owner', 'group', 'type'):
                attr += '_str'
            txt = sep*length if attr=='sep' else text2wrap(getattr(self, attr), length, ljust)
            ls.append(txt)
        return ''.join(ls)

    def __repr__(self):
        return '<FSEntry: {}>'.format(self.name)

    def get_type_from_ext(self, files_ext):
        if self.is_dir:
            return 'dir'
        elif self.type == FileType.exe:
            return 'exe'
        else:
            for ftype in files_ext.keys():
                if self.ext[1:] in files_ext[ftype]:
                    return ftype
            else:
                return 'reg'

    @property
    def is_dir(self):
        return self.type in (FileType.dir, FileType.link2dir)

    @property
    def is_link(self):
        return self.type in (FileType.link2dir, FileType.link, FileType.nlink)

    @property
    def is_dotfile(self):
        return self.name[0] == '.'

    def is_filtered(self, fs_hide, fs_show):
        """Returns if a file should be filtered (hidden) or not.
        Accepts 2 lists: first items to hide, second items to show.
        If a file is in both lists, it is shown"""
        if len(fs_show)==0 and len(fs_hide)==0:
            return False
        return False if self.pfile in fs_show else self.pfile in fs_hide

    def update_size(self, size):
        self.size = size
        self.size_str = size2str(size)


########################################################################
##### BaseFolder
class BaseFolder:
    """Base Folder"""

    def __init__(self, base, rel='', parfs=None):
        self.parfs = parfs
        self.cfg = FSConfig()
        adir = expanduser(base if VFS_STRING in base else normpath(base))
        if not isabs(adir):
            adir = abspath(adir)
        if rel != '':
            rel = normpath(rel)
        self.prepare_paths(adir, rel)
        self.pre_init()
        self.load()

    def __len__(self):
        return len(self._items_sorted)

    def __getitem__(self, key):
        return self._items_sorted.__getitem__(key)

    def __repr__(self):
        if self.vfs:
            return '<{}: {} [{}]>'.format(self.clsname, self.path_str, self.pdir)
        else:
            return '<{}: {}>'.format(self.clsname, self.path_str)

    def dump_info(self):
        log.info('##### {}'.format(self))
        log.info('##### base:  {} {}'.format(self.base, self.rel))
        log.info('##### rbase: {} {}'.format(self.rbase, self.rel))
        log.info('##### pdir:  {}'.format(self.pdir))
        log.info('##### parfs: {}'.format(self.parfs))

    @property
    def nfiltered(self):
        return len(self._items) - len(self._items_sorted) + 1 # pardir

    @property
    def path_str(self):
        return self.base + self.rel

    @property
    def base_filename(self):
        return self.base[:-len(VFS_STRING)] if self.base.endswith(VFS_STRING) else self.base

    @property
    def dirname(self):
        if self.path_str.endswith(VFS_STRING):
            if self.clsname == 'CompressedFileFolder':
                # root of vfsfile (compressedfile) -> dir of vfsfile
                parent = dirname(self.path_str[:-len(VFS_STRING)])
            else:
                # root of vfsfile (search) -> vfsfile
                parent = self.path_str[:-len(VFS_STRING)]
        else:
            parent = dirname(self.path_str)
            if parent.endswith('#vfs:'):
                parent += '//'
        return parent

    @property
    def basename(self):
        return basename(self.path_str[:-len(VFS_STRING)] if self.path_str.endswith(VFS_STRING) else self.path_str)

    def prepare_paths(adir, rel):
        """Method called to initiale paths.
        'adir' is str, 'rel' is str"""
        raise NotImplementedError

    def pre_init(self):
        """Method called after initializing paths but before load contents"""
        raise NotImplementedError

    def exit(self, all_levels=False, rebuild=False):
        """Method to be called at folder destruction"""
        raise NotImplementedError

    def chdir(self, rel):
        """Method called before changing directory"""
        raise NotImplementedError

    def load(self):
        """Load folder contents. Call when path contents have change"""
        log.debug('Load {} [{}]'.format(self, self.pdir))
        self._items = [FSEntry(join(self.pdir, f)) for f in os.listdir(self.pdir)]

    def refresh(self):
        """Apply filters and sort entries. Call when config, sorting or filters change"""
        log.debug('Refresh {} [{}]'.format(self, self.pdir))
        ds, fs = list(), list()
        if self.cfg.filters == '':
            fs_hide, fs_show = self.get_filtered('!.*,!*') # show all
        else:
            fs_hide, fs_show = self.get_filtered(self.cfg.filters)
        if self.cfg.sort_mix_dirs:
            fs = [it for it in self._items if (self.cfg.show_dotfiles and it.is_dotfile or not it.is_dotfile)
                                              and (not it.is_filtered(fs_hide, fs_show))]
        else:
            for it in self._items:
                if not self.cfg.show_dotfiles and it.is_dotfile:
                    continue
                if it.is_filtered(fs_hide, fs_show):
                    continue
                if it.is_dir:
                    ds.append(it)
                else:
                    fs.append(it)
        key = self.cfg.get_sort_key()
        ds.sort(key=key, reverse=self.cfg.sort_reverse)
        ds.insert(0, FSEntry(join(self.pdir, '..'))) # insert parent dir
        fs.sort(key=key, reverse=self.cfg.sort_reverse)
        ds.extend(fs)
        self._items_sorted = ds

    def lookup(self, filename):
        """Returns FSEntry if file exists in directory"""
        for f in self._items_sorted:
            if f.name == filename:
                return f
        else:
            return None

    def pos(self, filename):
        """Returns the position index if file exists in directory"""
        for f in self._items_sorted:
            if f.name == filename:
                return self._items_sorted.index(f)
        else:
            return -1

    def get_filenames(self, start=0):
        """Returns a (sub)list with the sorted dirs/files names in directory"""
        return [f.name for f in self._items_sorted[start:]]

    def get_filtered(self, filters):
        """Returns entries that match any of the globs.
        If glob starts with ! add match to 2nd list"""
        globs = [] if filters=='' else [f.strip() for f in filters.split(',')]
        fs_hide, fs_show = [], []
        p = escape(self.pdir)
        for g in globs:
            if g.startswith('!'):
                fs_show.extend(glob(join(p, g[1:])))
            else:
                fs_hide.extend(glob(join(p, g)))
        return fs_hide, fs_show

    @property
    def dirs(self):
        return [it for it in self._items_sorted if it.is_dir and it.name!=os.pardir]


########################################################################
##### LocalFolder
class LocalFolder(BaseFolder):
    """Local File System Folder"""

    clsname = 'LocalFolder'

    def prepare_paths(self, adir, _):
        assert isdir(adir)
        self.base = self.rbase = ''
        self.pdir = self.rel = adir
        self.vfs = False
        self.vfs_str = ''

    def pre_init(self):
        pass

    def exit(self, all_levels=False, rebuild=False):
        pass

    def chdir(self, rel):
        raise RuntimeError


########################################################################
##### CompressedFileFolder
class CompressedFileFolder(BaseFolder):
    """Compressed File Folder"""

    clsname = 'CompressedFileFolder'

    def prepare_paths(self, adir, rel):
        self.base = adir + VFS_STRING
        self.rbase = mkdtemp(suffix='.lfm')
        self.rel = rel
        self.pdir = join(self.rbase, rel)
        self.vfs = True
        self.vfs_str = VFS_STRING
        self.already_exited = False

    def pre_init(self):
        log.debug('Preparing VFS: {}'.format(self))
        c = get_compressed_file_engine(self.base.replace(VFS_STRING, ''))
        if c is None:
            raise
        st, res, err = ProcessCommand('Creating VFS', self.base_filename,
                                      c.cmd_uncompress, path=self.rbase).run()
        if st == -100: # stopped by user
            delete_bulk(self.rbase, ignore_errors=True)
            raise UserWarning('Stopped by user')
        if err != '':
            delete_bulk(self.rbase, ignore_errors=True)
            raise UserWarning(err)

    def exit(self, all_levels=False, rebuild=False):
        if self.already_exited:
            return
        log.debug('Exiting from VFS: {}'.format(self))
        if all_levels:
            parfs = self.parfs
            while parfs:
                parfs.exit()
                parfs = parfs.parfs
        if rebuild:
            self.__rebuild()
        delete_bulk(self.rbase, ignore_errors=True)
        self.already_exited = True

    def chdir(self, rel):
        self.rel = rel
        new_pdir = join(self.rbase, rel)
        if isdir(new_pdir):
            self.pdir = new_pdir
            log.debug('VFS: chdir {} -> {}'.format(self, self.pdir))
            self.load()
            return self
        else:
            log.debug('VFS in VFS: {} -> {}'.format(self, self.pdir))
            log.debug('New folder: path="{}", oldfs={}'.format(new_pdir, '"%s"' % self))
            base, rel = split_vfs_path(new_pdir)
            newfs = open_folder(base, rel, self, True)
            newfs.base = self.path_str + VFS_STRING
            return newfs

    def __rebuild(self):
        log.debug('Rebuilding VFS: {}'.format(self))
        c = get_compressed_file_engine(self.base.replace(VFS_STRING, ''))
        if c is None:
            raise
        _, tmpfile = mkstemp(suffix='.lfm')
        st, res, err = ProcessCommand('Rebuilding VFS', self.base_filename,
                                      c.cmd_compress2('*', tmpfile), path=self.rbase).run()
        # compress process always create filename with extension
        delete_bulk(tmpfile, ignore_errors=True)
        tmpfile_ext = tmpfile + c.exts[0]
        if st == -100: # stopped by user
            delete_bulk(tmpfile_ext, ignore_errors=True)
            delete_bulk(self.rbase, ignore_errors=True)
            raise UserWarning('Stopped by user')
        if err != '':
            delete_bulk(tmpfile_ext, ignore_errors=True)
            delete_bulk(self.rbase, ignore_errors=True)
            raise UserWarning(err)
        try:
            overwrite_vfsfile(tmpfile_ext, c.path)
        except OSError as err:
            delete_bulk(tmpfile_ext, ignore_errors=True)
            raise


########################################################################
##### SearchFolder
class SearchFolder(BaseFolder):
    """Search Folder"""

    clsname = 'SearchFolder'

    def prepare_paths(self, adir, rel):
        self.base = adir + VFS_STRING
        self.rbase = mkdtemp(suffix='.lfm')
        self.rel = rel
        self.pdir = join(self.rbase, rel)
        self.vfs = True
        self.vfs_str = VFS_STRING

    def pre_init(self):
        pass

    def copy_files(self, files):
        log.debug('Preparing VFS: {}'.format(self))
        self.files = files
        for f in files:
            src = join(self.parfs.pdir, f)
            dest = join(self.pdir, f)
            os.makedirs(dirname(dest), exist_ok=True)
            copy_bulk(src, dest)
        self.load()
        return self

    def exit(self, all_levels=False, rebuild=False):
        log.debug('Exiting from VFS: {}'.format(self))
        if rebuild:
            self.__rebuild()
        delete_bulk(self.rbase, ignore_errors=True)

    def chdir(self, rel):
        self.rel = rel
        self.pdir = join(self.rbase, rel)
        log.debug('Search VFS: chdir {} -> {}'.format(self, self.pdir))
        self.load()
        return self

    def __rebuild(self):
        log.debug('Rebuilding VFS: {}'.format(self))
        files = [f[2:] for f in run_in_cli('find . -type f', self.pdir)[0].split()]
        deleted = set(self.files) - set(files)
        for f in files:
            src = join(self.pdir, f)
            dest = join(self.parfs.pdir, f)
            os.makedirs(dirname(dest), exist_ok=True)
            copy_bulk(src, dest)
        for f in deleted:
            dest = join(self.parfs.pdir, f)
            delete_bulk(dest, ignore_errors=True)


########################################################################
def split_vfs_path(path):
    n = path.rfind(VFS_STRING) # can be vfs in vfs => rfind, not find!
    if n == -1:
        base, rel = path, ''
    else:
        base, rel = path[:n+len(VFS_STRING)], path[n+len(VFS_STRING):]
    return base, rel

def open_folder(base, rel, oldfs, vfsinvfs):
    if isdir(base):
        if rel != '':
            raise ValueError
        return LocalFolder(base)
    if base.endswith(VFS_STRING):
        base = base[:-len(VFS_STRING)]
    if check_compressed_vfs(base):
        if not vfsinvfs and oldfs and oldfs.parfs and exists(oldfs.parfs.pdir):
            oldfs.parfs.rel = dirname(oldfs.parfs.rel)
            return oldfs.parfs
        else:
            return CompressedFileFolder(base, rel, oldfs)
    raise FileNotFoundError

def is_delete_oldfs(newpath, oldfs):
    """Returns if we will exit oldfs in next new_folder()"""
    base, rel = split_vfs_path(newpath)
    return oldfs and oldfs.vfs and base!=oldfs.base and not basename(newpath)==basename(oldfs.path_str)

def new_folder(path, oldfs=None, rebuild_if_exit=False, files=None):
    # path is str, oldfs is BaseFolder!
    if files:
        log.debug('New searchvfs folder: path="{}", oldfs={}'.format(path, '"%s"' % oldfs if oldfs else 'none'))
        return SearchFolder(path, parfs=oldfs).copy_files(files)
    log.debug('New folder: path="{}", oldfs={}'.format(path, '"%s"' % oldfs if oldfs else 'none'))
    base, rel = split_vfs_path(path)
    if not oldfs:
        return open_folder(base, rel, None, False)
    if base == oldfs.base:
        return oldfs.chdir(rel)
    else:
        if basename(path) != basename(oldfs.path_str) or path.endswith(VFS_STRING) \
           and oldfs.path_str.endswith(VFS_STRING+VFS_STRING): # exit vfs
            oldfs.exit(rebuild=rebuild_if_exit)
            if oldfs.clsname == 'SearchFolder':
                return oldfs.parfs
        return open_folder(base, rel, oldfs, False)


########################################################################



#!/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2001-17  Iñigo Serna
# Time-stamp: <2017-06-25 18:31:04 inigo>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""lfm v3.0 - (C) 2001-17, by Iñigo Serna <inigoserna@gmail.com>

'Last File Manager' is a powerful file manager for UNIX console.
It has a curses interface and it's written in Python version 3.4+.
Released under GNU Public License, read COPYING file for more details.
"""

__author__ = 'Iñigo Serna'
__revision__ = '3.0'


import os
from os.path import join, exists
import sys
import argparse
import logging

from common import *
from ui import run_app


########################################################################
######################################################################
##### Main
def lfm_exit(ret_code, ret_path='.'):
    with open('/tmp/lfm-%s.path' % (os.getppid()), 'w') as f:
        f.write(ret_path)
    sys.exit(ret_code)


def helper_delete_item(filename, text):
    try:
        os.unlink(filename)
    except OSError as err:
        print('lfm - ERROR: {}\n{}'.format(text, err))
    else:
        print('lfm: {}!'.format(text))
    sys.exit(0)


def lfm_start():
    parser = argparse.ArgumentParser(description=__doc__.split('\n')[0],
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog='\n'.join(__doc__.split('\n')[2:]))
    parser.add_argument('-d', '--debug', help='Enable debug level in log file', action='store_true')
    parser.add_argument('-w', '--use-wide-chars', help='Enable support for wide chars', action='store_true')
    parser.add_argument('--restore-config', help='Restore default configuration', action='store_true')
    parser.add_argument('--restore-keys', help='Restore default key bindings', action='store_true')
    parser.add_argument('--restore-theme', help='Restore default theme', action='store_true')
    parser.add_argument('--delete-history', help='Delete history', action='store_true')
    parser.add_argument('path1', nargs='?', default='.', help='Path to show in left pane (default: ".")')
    parser.add_argument('path2', nargs='?', default='.', help='Path to show in right pane (default: ".")')
    args = parser.parse_args()

    if not exists(CONFIG_DIR):
        os.makedirs(CONFIG_DIR)
    logging.basicConfig(level=logging.DEBUG if args.debug else DEBUG_LEVEL,
                        # format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                        format='%(asctime)s  %(name)s(%(module)12s:%(lineno)4d)  %(levelname).1s  %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        filename=join(CONFIG_DIR, 'lfm.log'),
                        filemode='w')

    if args.restore_config:
        helper_delete_item(CONFIG_FILE, 'restore default configuration')
    if args.restore_keys:
        helper_delete_item(KEYS_FILE, 'restore default key bindings')
    if args.restore_theme:
        helper_delete_item(THEME_FILE, 'restore default theme')
    if args.delete_history:
        helper_delete_item(HISTORY_FILE, 'delete history')

    try:
        pwd = os.getcwd()
        os.chdir(args.path1)
        os.chdir(args.path2)
        os.chdir(pwd)
    except OSError as err:
        print('lfm - ERROR: cannot initialize path\n{}'.format(err))
        sys.exit(-1)

    logging.info('Start lfm')
    try:
        path = run_app(args.path1, args.path2, args.use_wide_chars)
    except FileNotFoundError as err:
        print('ERROR: Cannot copy default theme or keys file to user configuration folder {}\n{}\nQuitting'.format(CONFIG_DIR, err))
        sys.exit(-1)
    except BaseException as err:
        import traceback
        tb_str = traceback.format_exc()
        logging.critical('ERROR: ' + str(err))
        logging.critical(tb_str)
        print('ERROR:', str(err))
        print(tb_str)
        path = None
        raise

    logging.info('End lfm, returns: "{}"'.format(path))
    if path is not None:
        lfm_exit(0, path)
    else:
        lfm_exit(0)


########################################################################
if __name__ == '__main__':
    lfm_start()


########################################################################



# -*- coding: utf-8 -*-


import curses
import curses.panel
from os import sep, listdir
from os.path import basename, dirname, exists, expanduser, isdir, isabs, join

from utils import length, max_length, text2wrap, prev_step, next_step, perms2str, get_binary_programs, DirsTree
from common import *


########################################################################
##### Module variables
app = None
history = {}


########################################################################
##### Scrollbar
def display_scrollbar(win, y0, x0, h, n, i, a):
    """Display a scrollbar"""
    if n <= h:
        return
    win.vline(y0, x0, curses.ACS_VLINE, h)
    ss = max(h*h//n, 1)
    y = min(max((i//h)*h*h//n, 0), h-ss)
    win.vline(y0+y, x0, curses.ACS_CKBOARD, ss)
    if a != 0:
        win.vline(y0, x0, '^', 1)
        if (ss == 1) and (y == 0):
            win.vline(y0+1, x0, curses.ACS_CKBOARD, 1)
    if n > a + h:
        win.vline(y0+h-1, x0, 'v', 1)
        if (ss == 1) and (y == h-1):
            win.vline(y0+h-2, x0, curses.ACS_CKBOARD, 1)


######################################################################
##### Dialogs
def DialogMessage(title, subtitle):
    """Show a message. No wait, no keys"""
    title, subtitle = title[:app.w-14], subtitle[:app.w-14]
    h, w = 5, min(max(len(title), len(subtitle), 22)+6, app.w-2)
    try:
        win = curses.newwin(h, w, (app.h-h)//2, (app.w-w)//2)
        pwin = curses.panel.new_panel(win)
        pwin.top()
    except curses.error:
        raise
    win.keypad(1)
    win.bkgd(app.CLR['dialog'])
    win.erase()
    win.box()
    win.addstr(0, (w-len(title)-2)//2, ' %s ' % title, app.CLR['dialog_title'])
    win.addstr(2, 2, subtitle)
    win.addstr(h-1, (w-22)//2, ' Press Ctrl-C to stop ')
    win.refresh()


def DialogError(text):
    """Show an error message and waits for a key"""
    lines = text.split('\n')
    lth = max(max_length(lines), 31)
    h, w = min(len(lines)+4, app.h-2), min(lth+4, app.w-2)
    try:
        win = curses.newwin(h, w, (app.h-h)//2, (app.w-w)//2)
        pwin = curses.panel.new_panel(win)
        pwin.top()
    except curses.error:
        raise
    win.keypad(1)
    win.bkgd(app.CLR['dialog_error'])
    win.erase()
    win.box()
    win.addstr(0, (w-7)//2, ' Error ', app.CLR['dialog_error_title'])
    win.addstr(h-1, (w-27)//2, ' Press any key to continue ')
    for i, l in enumerate(lines):
        win.addstr(i+2, 2, l, app.CLR['dialog_error_text'])
    win.refresh()
    while not win.getch():
        pass
    pwin.hide()
    curses.panel.update_panels()


def DialogGetKey(title, question):
    """Show a message and return key pressed"""
    question = question.replace('\t', ' ' * 4)
    lines = question.split('\n')
    h = min(len(lines)+4, app.h-2)
    w = min(max_length(lines)+4, app.w-2)
    try:
        win = curses.newwin(h, w, (app.h-h)//2, (app.w-w)//2)
        pwin = curses.panel.new_panel(win)
        pwin.top()
    except curses.error:
        raise
    win.keypad(1)
    win.bkgd(app.CLR['dialog'])
    win.erase()
    win.box()
    win.addstr(0, (w-len(title)-2)//2, ' %s ' % title, app.CLR['dialog_title'])
    for i, l in enumerate(lines):
        win.addstr(i+2, 2, l)
    win.refresh()
    win.keypad(1)
    while True:
        ch = win.getch()
        if ch in (0x03, 0x1B): # Ctrl-C, ESC
            ch = -1
            break
        elif 0x01 <= ch <= 0xFF:
            break
        else:
            curses.beep()
    pwin.hide()
    curses.panel.update_panels()
    return ch


def DialogConfirm(title, question, default=0):
    """Show a yes/no question, returning 1/0"""
    BTN_SELECTED, BTN_NO_SELECTED = app.CLR['button_active'], app.CLR['button_inactive']
    h, w = 5, min(max(34, len(question)+5), app.w-2)
    try:
        win = curses.newwin(h, w, (app.h-h)//2, (app.w-w)//2)
        pwin = curses.panel.new_panel(win)
        pwin.top()
    except curses.error:
        raise
    win.keypad(1)
    win.bkgd(app.CLR['dialog'])
    win.erase()
    win.box()
    win.addstr(0, (w-len(title)-2)//2, ' %s ' % title, app.CLR['dialog_title'])
    win.addstr(1, 2, question)
    win.refresh()
    row, col = (app.h-h)//2 + 3, (app.w-w)//2
    col1, col2 = col + w//5 + 1, col + w*4//5 - 6
    win.keypad(1)
    answer = default
    while True:
        if answer == 1:
            attr_yes, attr_no = BTN_SELECTED, BTN_NO_SELECTED
        else:
            attr_yes, attr_no = BTN_NO_SELECTED, BTN_SELECTED
        btn = curses.newpad(1, 8)
        btn.addstr(0, 0, '[ Yes ]', attr_yes)
        btn.refresh(0, 0, row, col1, row+1, col1+6)
        btn = curses.newpad(1, 7)
        btn.addstr(0, 0, '[ No ]', attr_no)
        btn.refresh(0, 0, row, col2, row+1, col2+5)
        ch = win.getch()
        if ch in (curses.KEY_UP, curses.KEY_DOWN, curses.KEY_LEFT, curses.KEY_RIGHT, 9, curses.KEY_BTAB):
            answer = not answer
        elif ch in (ord('Y'), ord('y')):
            answer = 1
            break
        elif ch in (ord('N'), ord('n')):
            answer = 0
            break
        elif ch in (0x03, 0x1B):    # Ctrl-C, ESC
            answer = 0
            break
        elif ch in (10, 13):        # enter
            break
        else:
            curses.beep()
    pwin.hide()
    curses.panel.update_panels()
    return answer


def DialogConfirmAll(title, question, default=0):
    """Show a yes/all/no/stop question, returning 1/2/0/-1"""
    BTN_SELECTED, BTN_NO_SELECTED = app.CLR['button_active'], app.CLR['button_inactive']
    h, w = 5, min(max(48, len(question)+5), app.w-2)
    try:
        win = curses.newwin(h, w, (app.h-h)//2, (app.w-w)//2)
        pwin = curses.panel.new_panel(win)
        pwin.top()
    except curses.error:
        raise
    win.keypad(1)
    win.bkgd(app.CLR['dialog'])
    win.erase()
    win.box()
    win.addstr(0, (w-len(title)-2)//2, ' %s ' % title, app.CLR['dialog_title'])
    win.addstr(1, 2, question)
    win.refresh()
    row, col = (app.h-h)//2 + 3, (app.w-w)//2
    x = (w-28) // 5
    col1 = col + x + 1
    col2 = col1 + 7 + x
    col3 = col2 + 7 + x
    col4 = col3 + 6 + x
    win.keypad(1)
    answer = default
    order = [1, 2, 0, -1]
    while True:
        attr_yes = attr_all = attr_no = attr_skipall = BTN_NO_SELECTED
        if answer == 1:
            attr_yes = BTN_SELECTED
        elif answer == 2:
            attr_all = BTN_SELECTED
        elif answer == 0:
            attr_no = BTN_SELECTED
        else: # answer == -1:
            attr_skipall = BTN_SELECTED
        btn = curses.newpad(1, 8)
        btn.addstr(0, 0, '[ Yes ]', attr_yes)
        btn.refresh(0, 0, row, col1, row+1, col1+6)
        btn = curses.newpad(1, 8)
        btn.addstr(0, 0, '[ All ]', attr_all)
        btn.refresh(0, 0, row, col2, row+1, col2+6)
        btn = curses.newpad(1, 7)
        btn.addstr(0, 0, '[ No ]', attr_no)
        btn.refresh(0, 0, row, col3, row+1, col3+5)
        btn = curses.newpad(1, 15)
        btn.addstr(0, 0, '[ Stop ]', attr_skipall)
        btn.refresh(0, 0, row, col4, row+1, col4+7)
        ch = win.getch()
        if ch in (curses.KEY_UP, curses.KEY_LEFT, curses.KEY_BTAB):
            try:
                answer = order[order.index(answer) - 1]
            except IndexError:
                answer = order[len(order)]
        elif ch in (curses.KEY_DOWN, curses.KEY_RIGHT, 9):
            try:
                answer = order[order.index(answer) + 1]
            except IndexError:
                answer = order[0]
        elif ch in (ord('Y'), ord('y')):
            answer = 1
            break
        elif ch in (ord('A'), ord('a')):
            answer = 2
            break
        elif ch in (ord('N'), ord('n')):
            answer = 0
            break
        elif ch in (ord('S'), ord('s'), 0x03, 0x1B):    # Ctrl-C, ESC
            answer = -1
            break
        elif ch in (10, 13):        # enter
            break
        else:
            curses.beep()
    pwin.hide()
    curses.panel.update_panels()
    return answer


def DialogConfirmAllNone(title, question, default=0):
    """Show a yes/all/no/none/stop question, returning 1/2/0/-2/-1"""
    BTN_SELECTED, BTN_NO_SELECTED = app.CLR['button_active'], app.CLR['button_inactive']
    h, w = 5, min(max(50, len(question)+5), app.w-2)
    try:
        win = curses.newwin(h, w, (app.h-h)//2, (app.w-w)//2)
        pwin = curses.panel.new_panel(win)
        pwin.top()
    except curses.error:
        raise
    win.keypad(1)
    win.bkgd(app.CLR['dialog'])
    win.erase()
    win.box()
    win.addstr(0, (w-len(title)-2)//2, ' %s ' % title, app.CLR['dialog_title'])
    win.addstr(1, 2, question)
    win.refresh()
    row, col = (app.h-h)//2 + 3, (app.w-w)//2
    x = (w-36) // 6
    col1 = col + x + 1
    col2 = col1 + 7 + x
    col3 = col2 + 7 + x
    col4 = col3 + 6 + x
    col5 = col4 + 8 + x
    win.keypad(1)
    answer = default
    order = [1, 2, 0, -2, -1]
    while True:
        attr_yes = attr_all = attr_no = attr_none = attr_skipall = BTN_NO_SELECTED
        if answer == 1:
            attr_yes = BTN_SELECTED
        elif answer == 2:
            attr_all = BTN_SELECTED
        elif answer == 0:
            attr_no = BTN_SELECTED
        elif answer == -2:
            attr_none = BTN_SELECTED
        else: # answer == -1:
            attr_skipall = BTN_SELECTED
        btn = curses.newpad(1, 8)
        btn.addstr(0, 0, '[ Yes ]', attr_yes)
        btn.refresh(0, 0, row, col1, row+1, col1+6)
        btn = curses.newpad(1, 8)
        btn.addstr(0, 0, '[ All ]', attr_all)
        btn.refresh(0, 0, row, col2, row+1, col2+6)
        btn = curses.newpad(1, 7)
        btn.addstr(0, 0, '[ No ]', attr_no)
        btn.refresh(0, 0, row, col3, row+1, col3+5)
        btn = curses.newpad(1, 9)
        btn.addstr(0, 0, '[ NOne ]', attr_none)
        btn.refresh(0, 0, row, col4, row+1, col4+7)
        btn = curses.newpad(1, 9)
        btn.addstr(0, 0, '[ Stop ]', attr_skipall)
        btn.refresh(0, 0, row, col5, row+1, col5+7)
        ch = win.getch()
        if ch in (curses.KEY_UP, curses.KEY_LEFT, curses.KEY_BTAB):
            try:
                answer = order[order.index(answer) - 1]
            except IndexError:
                answer = order[len(order)]
        elif ch in (curses.KEY_DOWN, curses.KEY_RIGHT, 9):
            try:
                answer = order[order.index(answer) + 1]
            except IndexError:
                answer = order[0]
        elif ch in (ord('Y'), ord('y')):
            answer = 1
            break
        elif ch in (ord('A'), ord('a')):
            answer = 2
            break
        elif ch in (ord('N'), ord('n')):
            answer = 0
            break
        elif ch in (ord('O'), ord('o')):
            answer = -2
            break
        elif ch in (ord('S'), ord('s'), 0x03, 0x1B):    # Ctrl-C, ESC
            answer = -1
            break
        elif ch in (10, 13):        # enter
            break
        else:
            curses.beep()
    pwin.hide()
    curses.panel.update_panels()
    return answer


########################################################################
##### CursorAnimation
class CursorAnimation:
    """A small progress animation show on top-right corner of screen"""

    anim_chars = ('|', '/', '-', '\\')

    def __init__(self):
        self.step = 0
        self.win = curses.newpad(1, 2)
        self.win.bkgd(app.CLR['dialog_title'])

    def next(self):
        self.win.erase()
        self.win.addch(CursorAnimation.anim_chars[self.step%4])
        self.win.refresh(0, 0, 0, app.w-3, 1, app.w-2)
        self.step = 0 if self.step>3 else self.step+1


########################################################################
##### DialogMessagePane
class DialogMessagePanel:
    """A dialog to show a message. No wait"""

    def __init__(self, title, subtitle):
        title, subtitle = title[:app.w-14], subtitle[:app.w-14]
        h, w = 5, min(max(len(title), len(subtitle), 22)+6, app.w-2)
        try:
            win = curses.newwin(h, w, (app.h-h)//2, (app.w-w)//2)
            self.pwin = curses.panel.new_panel(win)
            self.pwin.top()
        except curses.error:
            raise
        win.keypad(1)
        win.nodelay(1)
        win.bkgd(app.CLR['dialog'])
        win.erase()
        win.box()
        win.addstr(0, (w-len(title)-2)//2, ' %s ' % title, app.CLR['dialog_title'])
        win.addstr(2, 2, subtitle)
        win.addstr(h-1, (w-22)//2, ' Press Ctrl-C to stop ')
        win.refresh()

    def show(self):
        self.pwin.show()
        curses.panel.update_panels()

    def hide(self):
        self.pwin.hide()
        curses.panel.update_panels()

    def check_key(self):
        return self.pwin.window().getch()


########################################################################
##### DialogProgressXPanel
class DialogProgress1Panel:
    """A dialog with 1 progress bar"""

    def __init__(self, title):
        title = title[:app.w-14]
        h, self.w = 7, min(max(len(title), 60)+6, app.w-2)
        try:
            self.win = curses.newwin(h, self.w, (app.h-h)//2, (app.w-self.w)//2)
            self.pwin = curses.panel.new_panel(self.win)
            self.pwin.top()
        except curses.error:
            raise
        self.win.keypad(1)
        self.win.nodelay(1)
        self.win.bkgd(app.CLR['dialog'])
        self.win.erase()
        self.win.box()
        self.win.addstr(0, (self.w-len(title)-2)//2, ' %s ' % title, app.CLR['dialog_title'])
        self.win.addstr(2, 2, 'File:')
        self.win.addstr(4, 2, ' '*(self.w-4-7), app.CLR['progressbar_bg'])
        self.win.addstr(4, self.w-8, '[  0%]')
        self.win.addstr(h-1, (self.w-22)//2, ' Press Ctrl-C to stop ')
        self.win.refresh()

    def show(self):
        self.pwin.show()
        curses.panel.update_panels()

    def hide(self):
        self.pwin.hide()
        curses.panel.update_panels()
        app.display()

    def check_key(self):
        return self.pwin.window().getch()

    def update(self, text, i, n):
        buf = '{}/{}'.format(i, n)
        self.win.addstr(2, 9, text2wrap(text, self.w-11-11, fill=True), app.CLR['dialog_title'])
        self.win.addstr(2, self.w-2-len(buf), buf)
        self.win.addstr(4, 2, ' '*int((self.w-4-7)*i/n), app.CLR['progressbar_fg'])
        self.win.addstr(4, self.w-8, '[{:3}%]'.format(100*i//n))


class DialogProgress2Panel(DialogProgress1Panel):
    """A dialog with 2 progress bar"""

    def __init__(self, title):
        title = title[:app.w-14]
        h, self.w = 8, min(max(len(title), 60)+6, app.w-2)
        try:
            self.win = curses.newwin(h, self.w, (app.h-h)//2, (app.w-self.w)//2)
            self.pwin = curses.panel.new_panel(self.win)
            self.pwin.top()
        except curses.error:
            raise
        self.win.keypad(1)
        self.win.nodelay(1)
        self.win.bkgd(app.CLR['dialog'])
        self.win.erase()
        self.win.box()
        self.win.addstr(0, (self.w-len(title)-2)//2, ' %s ' % title, app.CLR['dialog_title'])
        self.win.addstr(2, 2, 'File:')
        self.win.addstr(4, 2, 'Bytes')
        self.win.addstr(4, 9, ' '*(self.w-4-14), app.CLR['progressbar_bg'])
        self.win.addstr(4, self.w-8, '[  0%]')
        self.win.addstr(5, 2, 'Count')
        self.win.addstr(5, 9, ' '*(self.w-4-14), app.CLR['progressbar_bg'])
        self.win.addstr(5, self.w-8, '[  0%]')
        self.win.addstr(h-1, (self.w-22)//2, ' Press Ctrl-C to stop ')
        self.win.refresh()

    def update(self, text, i, n, sp, st):
        buf = '{}/{}'.format(i, n)
        self.win.addstr(2, 9, text2wrap(text, self.w-11-11, fill=True), app.CLR['dialog_title'])
        self.win.addstr(2, self.w-2-len(buf), buf)
        self.win.addstr(4, 9, ' '*int((self.w-4-14)*sp/st), app.CLR['progressbar_fg'])
        self.win.addstr(4, self.w-8, '[{:3}%]'.format(100*sp//st))
        self.win.addstr(5, 9, ' '*int((self.w-4-14)*i/n), app.CLR['progressbar_fg'])
        self.win.addstr(5, self.w-8, '[{:3}%]'.format(100*i//n))


########################################################################
##### SelectItem
class SelectItem:
    """A dialog to select an item in a list"""

    def __init__(self, title, entries, y0=-1, x0=-1, entry_i='', quick_key=True, min_height=False):
        self.quick_key = quick_key
        if y0==-1 and x0==-1: # (y0,x0) is the position, if == -1 => center
            self.h = min(app.h-4, len(entries) if min_height else 10) + 2
            self.w = min(app.w-12, max(max_length(entries), len(title), 32)) + 8
            y0, x0 = (app.h-self.h) // 2, (app.w-self.w) // 2
        else:
            self.h = min(app.h-y0-2, 10) + 2 # len(entries) => no, fixed height
            self.w = max(min(app.w-2*x0-4-10, max_length(entries)), len(title), 20) + 8
        try:
            self.win = curses.newwin(self.h, self.w, y0, x0)
            self.pwin = curses.panel.new_panel(self.win)
            self.pwin.top()
        except curses.error:
            raise
        self.win.keypad(1)
        self.win.bkgd(app.CLR['selectitem'])
        self.entries = entries
        self.nels = len(entries)
        try:
            self.entry_i = self.entries.index(entry_i)
        except:
            self.entry_i = 0
        self.title = title

    def show(self):
        self.win.erase()
        self.win.box()
        h0, w0 = self.h-2, self.w-4
        if self.title != '':
            self.win.addstr(0, (self.w-len(self.title)-2)//2, ' %s ' % self.title, app.CLR['selectitem_title'])
        entry_a = self.entry_i//h0 * h0
        for i in range(h0):
            if entry_a+i >= self.nels:
                break
            line = text2wrap(self.entries[entry_a+i], w0)
            if entry_a+i == self.entry_i:
                self.win.addstr(i+1, 2, line, app.CLR['selectitem_cursor'])
            else:
                self.win.addstr(i+1, 2, line)
        self.win.refresh()
        display_scrollbar(self.win, 1, self.w-1, h0, self.nels, self.entry_i, entry_a)

    def manage_keys(self):
        initials = {ord(e[0]) for e in self.entries}
        while True:
            self.show()
            ch = self.win.getch()
            if self.quick_key:
                if ch in initials:
                    for e in self.entries:
                        if ch == ord(e[0]):
                            return self.entries[self.entries.index(e)]
            if ch in (0x03, 0x1B, ord('q'), ord('Q')):     # Ctrl-C, ESC
                return -1
            elif ch in (curses.KEY_UP, ord('k'), ord('K')):
                self.entry_i = max(0, self.entry_i-1)
            elif ch in (curses.KEY_DOWN, ord('j'), ord('J')):
                self.entry_i = min(self.entry_i+1, self.nels-1)
            elif ch in (curses.KEY_PPAGE, curses.KEY_BACKSPACE, 0x08, 0x02):
                self.entry_i = max(0, self.entry_i-(self.h-2))
            elif ch in (curses.KEY_NPAGE, ord(' '), 0x06):
                self.entry_i = min(self.entry_i+(self.h-2), self.nels-1)
            elif ch in (curses.KEY_HOME, 0x01):
                self.entry_i = 0
            elif ch in (curses.KEY_END, 0x05):
                self.entry_i = self.nels - 1
            elif ch == 0x0C:     # Ctrl-L
                entry_a = int(self.entry_i//(self.h-2)) * (self.h-2)
                self.entry_i = entry_a + (self.h-2)//2
            elif ch == 0x13:     # Ctrl-S
                ch2 = self.win.getkey()
                for e in self.entries[self.entry_i:]:
                    if e.find(ch2) == 0:
                        self.entry_i = self.entries.index(e)
                        break
            elif ch in (0x0A, 0x0D):   # enter
                return self.entries[self.entry_i]
            else:
                curses.beep()

    def run(self):
        selected = self.manage_keys()
        self.pwin.hide()
        curses.panel.update_panels()
        app.display()
        return selected


######################################################################
##### DialogFindGrep
class DialogFindGrep:
    """A dialog similar to SelectItem"""

    def __init__(self, title, entries, entry_i=''):
        self.h = min(app.h-8, len(entries), 12) + 5
        self.w = min(app.w-6, 100, max(max_length(entries), 60, len(title))) + 4
        y0, x0 = (app.h-self.h) // 2, (app.w-self.w) // 2
        try:
            self.win = curses.newwin(self.h, self.w, y0, x0)
            self.pwin = curses.panel.new_panel(self.win)
            self.pwin.top()
        except curses.error:
            raise
        self.win.keypad(1)
        self.win.bkgd(app.CLR['selectitem'])
        self.entries = entries
        self.nels = len(entries)
        try:
            self.entry_i = self.entries.index(entry_i)
        except:
            self.entry_i = 0
        self.title = title
        self.btn_active = 0

    def show(self):
        BTN_SELECTED, BTN_NO_SELECTED = app.CLR['selectitem_cursor'], app.CLR['selectitem']
        self.win.erase()
        self.win.box()
        h0, w0 = self.h-4, self.w-4
        if self.title != '':
            self.win.addstr(0, (self.w-len(self.title)-2)//2, ' %s ' % self.title, app.CLR['selectitem_title'])
        entry_a = self.entry_i//h0 * h0
        for i in range(h0):
            if entry_a+i >= self.nels:
                break
            line = text2wrap(self.entries[entry_a+i], w0, start_pct=0.9)
            if entry_a+i == self.entry_i:
                self.win.addstr(i+1, 2, line, app.CLR['selectitem_cursor'])
            else:
                self.win.addstr(i+1, 2, line)
        display_scrollbar(self.win, 1, self.w-1, h0, self.nels, self.entry_i, entry_a)
        self.win.hline(self.h-3, 1, curses.ACS_HLINE, self.w-2)
        self.win.hline(self.h-3, 0, curses.ACS_LTEE, 1)
        self.win.hline(self.h-3, self.w-1, curses.ACS_RTEE, 1)
        self.win.addstr(self.h-2, 3, '[ Go ]  [ Panelize ]  [ View ]  [ Edit ]  [ Do ]  [ Quit ]', BTN_NO_SELECTED)
        attrs = [BTN_NO_SELECTED] * 6
        attrs[self.btn_active] = BTN_SELECTED
        self.win.addstr(self.h-2, 3, '[ Go ]', attrs[0])
        self.win.addstr(self.h-2, 11, '[ PAnelize ]', attrs[1])
        self.win.addstr(self.h-2, 25, '[ View ]', attrs[2])
        self.win.addstr(self.h-2, 35, '[ Edit ]', attrs[3])
        self.win.addstr(self.h-2, 45, '[ Do ]', attrs[4])
        self.win.addstr(self.h-2, 53, '[ Quit ]', attrs[5])
        self.win.refresh()

    def manage_keys(self):
        while True:
            self.show()
            ch = self.win.getch()
            if ch in (0x03, 0x1B, ord('q'), ord('Q')):     # Ctrl-C, ESC
                return -1, self.entries[self.entry_i]
            elif ch in (curses.KEY_UP, ord('k'), ord('K')):
                self.entry_i = max(0, self.entry_i-1)
            elif ch in (curses.KEY_DOWN, ord('j'), ord('J')):
                self.entry_i = min(self.entry_i+1, self.nels-1)
            elif ch in (curses.KEY_PPAGE, curses.KEY_BACKSPACE, 0x08, 0x02):
                self.entry_i = max(0, self.entry_i-(self.h-2))
            elif ch in (curses.KEY_NPAGE, ord(' '), 0x06):
                self.entry_i = min(self.entry_i+(self.h-2), self.nels-1)
            elif ch in (curses.KEY_HOME, 0x01):
                self.entry_i = 0
            elif ch in (curses.KEY_END, 0x05):
                self.entry_i = self.nels - 1
            elif ch == 0x0C:     # Ctrl-L
                entry_a = int(self.entry_i//(self.h-2)) * (self.h-2)
                self.entry_i = entry_a + (self.h-2)//2
            elif ch == 0x13:     # Ctrl-S
                ch2 = self.win.getkey()
                for e in self.entries[self.entry_i:]:
                    if e.find(ch2) == 0:
                        self.entry_i = self.entries.index(e)
                        break
            elif ch in (curses.KEY_LEFT, curses.KEY_BTAB):
                self.btn_active = 5 if self.btn_active==0 else self.btn_active-1
            elif ch in (curses.KEY_RIGHT, 0x09): # tab
                self.btn_active = 0 if self.btn_active==5 else self.btn_active+1
            elif ch in (0x0A, 0x0D):   # enter
                return -1 if self.btn_active==5 else self.btn_active, self.entries[self.entry_i]
            elif ch in (ord('a'), ord('A')):
                return 1, self.entries[self.entry_i]
            elif ch in (curses.KEY_F3, ord('v'), ord('V')):
                return 2, self.entries[self.entry_i]
            elif ch in (curses.KEY_F4, ord('e'), ord('E')):
                return 3, self.entries[self.entry_i]
            elif ch in (ord('@'), ord('d'), ord('D')):
                return 4, self.entries[self.entry_i]
            else:
                curses.beep()

    def run(self):
        selected = self.manage_keys()
        self.pwin.hide()
        curses.panel.update_panels()
        app.display()
        return selected


######################################################################
##### Helper widgets
class Yes_No_Buttons:
    """Yes/No buttons"""

    def __init__(self, w, h, d):
        self.row = (app.h-h)//2 + 4 + d
        col = (app.w-w)//2
        self.col1, self.col2 = col + w//5 + 1, col + w*4//5 - 6
        self.active = 0

    def show(self):
        BTN_SELECTED = app.CLR['button_active']
        BTN_NO_SELECTED = app.CLR['button_inactive']
        if self.active == 0:
            attr1, attr2 = BTN_NO_SELECTED, BTN_NO_SELECTED
        elif self.active == 1:
            attr1, attr2 = BTN_SELECTED, BTN_NO_SELECTED
        else:
            attr1, attr2 = BTN_NO_SELECTED, BTN_SELECTED
        btn = curses.newpad(1, 8)
        btn.addstr(0, 0, '[<Yes>]', attr1)
        btn.refresh(0, 0, self.row, self.col1, self.row + 1, self.col1 + 6)
        btn = curses.newpad(1, 7)
        btn.addstr(0, 0, '[ No ]', attr2)
        btn.refresh(0, 0, self.row, self.col2, self.row + 1, self.col2 + 5)

    def manage_keys(self):
        tmp = curses.newpad(1, 1)
        tmp.keypad(1)
        while True:
            ch = tmp.getch()
            if ch in (0x03, 0x1B):      # Ctrl-C, ESC
                return -1
            elif ch in (ord('\t'), curses.KEY_BTAB):
                return ch
            elif ch in (10, 13):        # enter
                if self.active == 1:
                    return 10
                else:
                    return -1
            else:
                curses.beep()


######################################################################
##### EntryLine
class EntryLine:
    """An entry line to enter a dir, a file, a pattern, etc"""

    def __init__(self, par_widget, w, y0, x0, text='', history=None, is_files=True, cli=False):
        try:
            self.win = curses.newwin(1, w+1, y0, x0)
        except curses.error:
            raise
        self.par_widget = par_widget
        self.color = app.CLR['powercli_text'] if cli else app.CLR['entryline']
        self.win.attrset(self.color)
        self.win.keypad(1)
        self.win.nodelay(0)
        self.x0, self.y0, self.w = x0, y0, w
        self.text, self.origtext, self.pos, self.ins = text, text, len(text), True
        self.history, self.history_i = history, -1 if history is None else len(history)
        self.cli = cli
        self.is_files = True if self.cli else is_files

    def show(self):
        text, pos, w, ltext = self.text, self.pos, self.w, len(self.text)
        if pos < w:
            relpos = pos
            textstr = text[:w] if ltext>w else text.ljust(w)
        else:
            if pos > ltext - (w-1):
                relpos = w-1 - (ltext-pos)
                textstr = text[ltext-w+1:] + ' '
            else:
                relpos = pos - pos//w*w
                textstr = text[pos//w*w:pos//w*w+w]
        # self.win.refresh()  # needed to avoid a problem with blank paths
        self.win.bkgd(app.CLR['dialog'])
        self.win.erase()
        self.win.addstr(textstr[:w], self.color)
        self.win.move(0, relpos)
        self.win.refresh()

    def __select_item(self, entries, pos0, title=''):
        if not entries:
            curses.beep()
            return
        elif len(entries) == 1:
            return entries.pop()
        else:
            x = self.x0+pos0 if self.x0+pos0<3*app.w//4-4 else self.x0+2
            if self.cli:
                y = app.h-14 # SelectItem has min 12 height, else: app.h//2-2
            else:
                y = self.y0
            curses.curs_set(0)
            selected = SelectItem(title, entries, y+1, x-2, quick_key=False).run()
            app.display()
            if not self.cli:
                self.par_widget.show()
            curses.curs_set(1)
            return selected

    def __get_list_completion(self, text):
        tab_path = app.pane_active.tab_active.fs.pdir
        path = expanduser(text)
        path = path if isabs(path) else join(tab_path, path)
        try:
            if text.endswith(os.sep) and isdir(path):
                basedir, fs = path, listdir(path)
            else:
                basedir, start = dirname(path), basename(path)
                fs = [f for f in listdir(basedir) if f.startswith(start)]
        except OSError:
            return basedir, list()
        # sort files with dirs first
        d1, d2 = list(), list()
        for f in fs:
            if isdir(join(basedir, f)):
                d1.append(f+sep)
            else:
                d2.append(f)
        d1.sort()
        d1.extend(sorted(d2))
        return basedir, d1

    def manage_keys(self):
        while True:
            self.show()
            wch = self.win.get_wch()
            if isinstance(wch, str) and ord(wch)>=32 and ord(wch)!=127:
                if self.ins:
                    self.text = self.text[:self.pos] + wch + self.text[self.pos:]
                else:
                    self.text = self.text[:self.pos] + wch + self.text[self.pos+1:]
                self.pos += 1
            else: # int!
                wch = ord(wch) if isinstance(wch, str) else wch
                if wch in (3, 21):                       # C-c, ESC
                    return -1
                if wch == 24 and self.cli:               # C-x
                    return -2
                elif wch in (9, curses.KEY_BTAB) and not self.cli: # tab, S-tab
                    return wch
                elif wch in (10, 13):                    # enter
                    return 10
                # movement
                elif wch in (curses.KEY_HOME, 1):        # home, C-a
                    self.pos = 0
                elif wch in (curses.KEY_END, 5):         # end, C-e
                    self.pos = len(self.text)
                elif wch in (curses.KEY_LEFT, 2):        # left, C-b
                    if self.pos > 0:
                        self.pos -= 1
                elif wch in (curses.KEY_RIGHT, 6):       # right, C-f
                    if self.pos < len(self.text):
                        self.pos += 1
                elif wch in (16, 0x222, 0x223, 0x224):                 # C-p, C-left
                    self.pos = prev_step(self.text, self.pos)
                elif wch in (14, 0x231, 0x232, 0x233):                 # C-n, C-right
                    self.pos = next_step(self.text, self.pos)
                # deletion
                elif wch in (127, curses.KEY_BACKSPACE): # Backspace
                    if len(self.text) > 0 and self.pos > 0:
                        self.text = self.text[:self.pos-1] + self.text[self.pos:]
                        self.pos -= 1
                elif wch in (17, ):                      # C-q, C-Backspace
                    pos = prev_step(self.text, self.pos)
                    self.text = self.text[:pos] + self.text[self.pos:]
                    self.pos = pos
                elif wch == curses.KEY_DC:               # Del
                    if self.pos < len(self.text):
                        self.text = self.text[:self.pos] + self.text[self.pos+1:]
                elif wch in (18, 0x208, 0x209):                 # C-r, C-Del
                    pos = next_step(self.text, self.pos)
                    if pos>0 and self.text[pos-1] in '.([{<"\'':
                        pos = max(0, pos-1)
                    if pos+1<len(self.text) and self.text[pos] in ' ':
                        pos += 1
                    self.text = self.text[:self.pos] + self.text[pos:]
                elif wch == 23:                          # C-w
                    self.text, self.pos = '', 0
                elif wch == 8:                           # C-h
                    self.text, self.pos = self.text[self.pos:], 0
                elif wch == 11:                          # C-k
                    self.text = self.text[:self.pos]
                # insertion
                elif wch == 26:                          # C-z
                    self.text, self.pos = self.origtext, len(self.origtext)
                elif wch == 22:                          # C-v
                    if self.is_files:
                        buf = app.pane_active.tab_active.current_filename
                        buf = '"{}"'.format(buf) if self.cli else buf
                        self.text = self.text[:self.pos] + buf + self.text[self.pos:]
                        self.pos += len(buf)
                elif wch == 19:                          # C-s
                    if self.is_files:
                        buf = app.pane_active.tab_active.dirname + sep
                        buf = '"{}"'.format(buf) if self.cli else buf
                        self.text = self.text[:self.pos] + buf + self.text[self.pos:]
                        self.pos += len(buf)
                elif wch == 15:                          # C-o
                    if self.is_files:
                        buf = app.pane_inactive.tab_active.dirname + sep
                        buf = '"{}"'.format(buf) if self.cli else buf
                        self.text = self.text[:self.pos] + buf + self.text[self.pos:]
                        self.pos += len(buf)
                elif wch in (4, 28):                     # C-d, C-\
                    if self.is_files:
                        bmks = [app.cfg.bookmarks[k] for k in sorted(app.cfg.bookmarks.keys())]
                        selected = self.__select_item(bmks, self.pos if self.cli else 0, 'Bookmarks')
                        if selected not in (None, -1, ''):
                            if self.cli:
                                self.text = self.text[:self.pos] + '"' + selected + '"' + self.text[self.pos:]
                                self.pos += len(selected)+2
                            else:
                                self.text, self.pos = selected, len(selected)
                elif wch == 25:                          # C-y
                    if self.is_files:
                        items = app.pane_active.tab_active.history[::-1]
                        items += [p for p in app.pane_inactive.tab_active.history[::-1] if p not in items]
                        if items:
                            selected = self.__select_item(items, self.pos if self.cli else 0, 'Previous paths')
                            if selected not in (None, -1, ''):
                                if self.cli:
                                    self.text = self.text[:self.pos] + '"' + selected + '"' + self.text[self.pos:]
                                    self.pos += len(selected)+2
                                else:
                                    self.text, self.pos = selected, len(selected)
                elif wch == 20 and not self.cli:         # C-t
                    if self.is_files:
                        basedir, entries = self.__get_list_completion(self.text)
                        selected = self.__select_item(entries, self.pos, 'Complete')
                        if selected not in (None, -1, ''):
                            basepath = app.pane_active.tab_active.fs.pdir + os.sep
                            self.text = join(basedir, selected)
                            if self.text.startswith(basepath):
                                self.text = join(basedir, selected)[len(basepath):]
                            self.pos = len(self.text)
                elif wch in (9, 20) and self.cli:        # tab, C-t
                    if self.text.rfind(' "', 0, self.pos) != -1:
                        pos1 = self.pos-1 if self.text[self.pos-1]=='"' else self.pos
                        pos0 = self.text.rfind(' "', 0, pos1) + 2
                    else:
                        pos0 = self.text.rfind(' ', 0, self.pos) + 1
                        pos1 = self.pos
                    text = self.text[pos0:pos1]
                    if pos0 == 0:
                        basedir, entries = None, [p for p in get_binary_programs() if p.startswith(text)]
                    else:
                        basedir, entries = self.__get_list_completion(text)
                    selected = self.__select_item(entries, pos0, 'Complete')
                    if selected is not None and selected != -1:
                        if basedir is None:
                            selected += ' '
                        else:
                            if basedir != app.pane_active.tab_active.fs.pdir:
                                selected = join(basedir, selected)
                        self.text = self.text[:pos0] + selected + self.text[pos1:]
                        self.pos = len(self.text[:pos0]+selected)
                # history
                elif wch == curses.KEY_UP:               # up
                    if self.history is not None and self.history_i > 0:
                        if self.history_i == len(self.history):
                            self.history.append(self.text)
                        self.history_i -= 1
                        self.text = self.history[self.history_i]
                        self.pos = len(self.text)
                elif wch == curses.KEY_DOWN:             # down
                    if self.history is not None and self.history_i < len(self.history)-1:
                            self.history_i += 1
                            self.text = self.history[self.history_i]
                            self.pos = len(self.text)
                elif wch == 7:                           # C-g
                    if self.cli:
                        BOOKMARKS_STR, HISTORY_STR = '----- Stored:  -----', '----- History: -----'
                        entries = [BOOKMARKS_STR]
                        entries.extend([c for c in app.cfg.powercli_favs if c.strip()])
                        entries.append(HISTORY_STR)
                        entries.extend([c for c in self.history if c.strip()])
                        selected = self.__select_item(entries, 0, 'History')
                        if selected not in (None, -1, '', BOOKMARKS_STR, HISTORY_STR):
                            self.text, self.pos = selected, len(selected)
                    else:
                        if self.history is not None and len(self.history) > 0:
                            selected = self.__select_item(self.history, 0, 'History')
                            if selected not in (None, -1):
                                self.text, self.pos = selected, len(selected)
                # other
                elif wch == curses.KEY_IC:               # insert
                    self.ins = not self.ins
                else:
                    curses.beep()


######################################################################
##### DialogEntry
class DialogEntry:
    """An entry dialog to enter a dir, a file, a pattern…"""

    def __init__(self, title, help, text='', history=None, is_files=True):
        h, w = 6, max(len(help)+5, app.w//2)
        try:
            self.win = curses.newwin(h, w, (app.h-h)//2, (app.w-w)//2)
            self.entry = EntryLine(self, w-4, (app.h-h)//2+2, (app.w-w+4)//2, text,
                                   history=history, is_files=is_files, cli=False)
            self.btns = Yes_No_Buttons(w, h, 0)
            self.pwin = curses.panel.new_panel(self.win)
            self.pwin.top()
        except curses.error:
            raise
        self.w, self.title, self.help = w, title, help
        self.win.keypad(1)
        self.win.bkgd(app.CLR['dialog'])
        self.active_widget = self.entry

    def show(self):
        self.win.erase()
        self.win.box()
        self.win.addstr(0, (self.w-len(self.title)-2)//2, ' %s ' % self.title, app.CLR['dialog_title'])
        self.win.addstr(1, 2, '%s:' % self.help)
        self.win.refresh()
        self.entry.show()
        self.btns.show()

    def run(self):
        self.show()
        curses.curs_set(1)
        quit = False
        while not quit:
            self.btns.show()
            ans = self.active_widget.manage_keys()
            if ans == -1:              # Ctrl-C
                quit = True
                answer = None
            elif ans == ord('\t'):     # tab
                if self.active_widget == self.entry:
                    self.active_widget = self.btns
                    self.btns.active = 1
                    curses.curs_set(0)
                    answer = self.entry.text
                elif self.active_widget == self.btns and self.btns.active == 1:
                    self.btns.active = 2
                    curses.curs_set(0)
                    answer = None
                else:
                    self.active_widget = self.entry
                    self.btns.active = 0
                    curses.curs_set(1)
            elif ans == curses.KEY_BTAB:     # S+tab
                if self.active_widget == self.entry:
                    self.active_widget = self.btns
                    self.btns.active = 2
                    curses.curs_set(0)
                    answer = None
                elif self.active_widget == self.btns and self.btns.active == 1:
                    self.active_widget = self.entry
                    self.btns.active = 0
                    curses.curs_set(1)
                else:
                    self.active_widget = self.btns
                    self.btns.active = 1
                    curses.curs_set(0)
                    answer = self.entry.text
            elif ans == 10:              # return values
                quit = True
                answer = self.entry.text.strip()
        curses.curs_set(0)
        self.pwin.hide()
        curses.panel.update_panels()
        app.display()
        return answer


class DialogDoubleEntry:
    """A dialog with 2 entries"""

    def __init__(self, title, help1, help2, text1, text2='', history1=None, history2=None, is_files=True):
        h, w = 9, max(len(help2)+5, app.w//2)
        try:
            self.win = curses.newwin(h, w, (app.h-h)//2, (app.w-w)//2)
            self.entry1 = EntryLine(self, w-4, (app.h-h)//2+2, (app.w-w+4)//2, text1,
                                    history=history1, is_files=is_files, cli=False)
            self.entry2 = EntryLine(self, w-4, (app.h-h)//2+5, (app.w-w+4)//2, text2,
                                    history=history2, is_files=is_files, cli=False)
            self.btns = Yes_No_Buttons(w, h, 3)
            self.pwin = curses.panel.new_panel(self.win)
            self.pwin.top()
        except curses.error:
            raise
        self.w, self.title, self.help1, self.help2 = w, title, help1, help2
        self.win.keypad(1)
        self.win.bkgd(app.CLR['dialog'])
        self.active_widget = self.entry1

    def show(self):
        self.win.erase()
        self.win.box()
        self.win.addstr(0, (self.w-len(self.title)-2)//2, ' %s ' % self.title, app.CLR['dialog_title'])
        self.win.addstr(1, 2, '%s:' % self.help1)
        self.win.addstr(4, 2, '%s:' % self.help2)
        self.win.refresh()
        self.entry1.show()
        self.entry2.show()
        self.btns.show()

    def run(self):
        self.show()
        curses.curs_set(1)
        quit = False
        while not quit:
            self.btns.show()
            ans = self.active_widget.manage_keys()
            if ans == -1:              # Ctrl-C
                quit = True
                answer = None
            elif ans == ord('\t'):     # tab
                if self.active_widget == self.entry1:
                    self.active_widget = self.entry2
                elif self.active_widget == self.entry2:
                    self.active_widget = self.btns
                    self.btns.active = 1
                    curses.curs_set(0)
                    answer = self.entry1.text, self.entry2.text
                elif self.active_widget == self.btns and self.btns.active == 1:
                    self.btns.active = 2
                    answer = None
                else:
                    self.active_widget = self.entry1
                    self.btns.active = 0
                    curses.curs_set(1)
            elif ans == curses.KEY_BTAB:     # S+tab
                if self.active_widget == self.entry1:
                    self.active_widget = self.btns
                    self.btns.active = 2
                    curses.curs_set(0)
                    answer = None
                elif self.active_widget == self.entry2:
                    self.active_widget = self.entry1
                elif self.active_widget == self.btns and self.btns.active == 1:
                    self.active_widget = self.entry2
                    self.btns.active = 0
                    curses.curs_set(1)
                    answer = self.entry1.text, self.entry2.text
                else:
                    self.btns.active = 1
                    answer = self.entry1.text, self.entry2.text
            elif ans == 10:              # return values
                quit = True
                answer = self.entry1.text.strip(), self.entry2.text.strip()
        curses.curs_set(0)
        self.pwin.hide()
        curses.panel.update_panels()
        app.display()
        return answer


########################################################################
##### DialogPerms & DialogOwner
class DialogPerms:
    """Dialog to change file permissions"""

    def __init__(self, filename, perms, i=0, n=0):
        self.h, self.w = 7+4, 42+4
        x0, y0 = int((app.w-self.w)/2), int((app.h-self.h)/2)
        try:
            self.win = curses.newwin(self.h, self.w, y0, x0)
            self.pwin = curses.panel.new_panel(self.win)
            self.pwin.top()
        except curses.error:
            raise
        self.win.keypad(1)
        self.win.bkgd(app.CLR['dialog'])
        self.file = filename
        self.perms_old = perms2str(perms)
        self.perms = [l for l in self.perms_old]
        self.recursive = False
        self.i, self.n, self.entry_i = i, n, 0

    def show_btns(self):
        attr_sel, attr_no = app.CLR['button_active'], app.CLR['button_inactive']
        self.win.addstr(self.h-2, self.w-21, '[<Ok>]', attr_sel if self.entry_i==11 else attr_no)
        self.win.addstr(self.h-2, self.w-13, '[ Cancel ]', attr_sel if self.entry_i==12 else attr_no)
        if self.n > 1:
            self.win.addstr(self.h-2, 3, '[ All ]', attr_sel if self.entry_i==13 else attr_no)
            self.win.addstr(self.h-2, 12, '[ Ignore ]', attr_sel if self.entry_i==14 else attr_no)

    def show(self):
        self.win.erase()
        self.win.box()
        attr, attr_sel, attr_no = app.CLR['dialog_title'], app.CLR['dialog_perms'], app.CLR['dialog']
        title = 'Change file(s) permissions'
        self.win.addstr(0, int((self.w-len(title)-2)/2), ' {} '.format(title), attr)
        if self.n > 1:
            buf = '{}/{}'.format(self.i, self.n)
            lbuf = len(buf)
            self.win.addstr(2, self.w-2-len(buf), buf)
        else:
            lbuf = 0
        self.win.addstr(2, 2, 'File:')
        self.win.addstr(2, 8, text2wrap(self.file, self.w-11-lbuf, fill=True), attr)
        self.win.addstr(4, 2, '      owner  group  other        recursive')
        self.win.addstr(5, 2, 'new:  [---]  [---]  [---]           [ ]')
        self.win.addstr(6, 2, 'old:  [---]  [---]  [---]           [ ]')
        perms = ''.join(self.perms)
        self.win.addstr(5, 9, perms[0:3], attr_sel if self.entry_i==0 else attr_no)
        self.win.addstr(5, 16, perms[3:6], attr_sel if self.entry_i==1 else attr_no)
        self.win.addstr(5, 23, perms[6:9], attr_sel if self.entry_i==2 else attr_no)
        self.win.addstr(5, 39, 'X' if self.recursive else ' ', attr_sel if self.entry_i==3 else attr_no)
        self.win.addstr(6, 9, self.perms_old[0:3])
        self.win.addstr(6, 16, self.perms_old[3:6])
        self.win.addstr(6, 23, self.perms_old[6:9])
        self.show_btns()
        self.win.refresh()

    def manage_keys(self):
        order = [0, 1, 2, 3, 13, 14, 11, 12] if self.n>1 else [0, 1, 2, 3, 11, 12]
        nels = len(order)
        while True:
            self.show()
            ch = self.win.getch()
            if ch in (0x03, 0x1B, ord('q'), ord('Q')):
                return -1, False, False
            elif ch in (ord('\t'), 0x09, curses.KEY_DOWN, curses.KEY_RIGHT):
                i = order.index(self.entry_i)
                self.entry_i = order[0 if i==nels-1 else i+1]
            elif ch in (curses.KEY_UP, curses.KEY_LEFT, curses.KEY_BTAB):
                i = order.index(self.entry_i)
                self.entry_i = order[nels-1 if i==0 else i-1]
            elif ch in (ord('r'), ord('R')):
                if 0 <= self.entry_i <= 2:
                    d = self.entry_i * 3
                    self.perms[d] = 'r' if self.perms[d]=='-' else '-'
            elif ch in (ord('w'), ord('W')):
                if 0 <= self.entry_i <= 2:
                    d = 1 + self.entry_i * 3
                    self.perms[d] = 'w' if self.perms[d]=='-' else '-'
            elif ch in (ord('x'), ord('X')):
                if 0 <= self.entry_i <= 2:
                    d = 2 + self.entry_i * 3
                    self.perms[d] = 'x' if self.perms[d]=='-' else '-'
            elif ch in (ord('t'), ord('T')):
                if self.entry_i == 2:
                    self.perms[8] = 't' if self.perms[8]=='-' else '-'
            elif ch in (ord('s'), ord('S')):
                if 0 <= self.entry_i <= 1:
                    d = 2 + self.entry_i * 3
                    self.perms[d] = 's' if self.perms[d]=='-' else '-'
            elif ch in (ord(' '), 0x0A, 0x0D):
                if self.entry_i == 3:
                    self.recursive = not self.recursive
                elif self.entry_i == 12:
                    return -1, False, False
                elif self.n>1 and self.entry_i==13:
                    return self.perms, self.recursive, True
                elif self.n>1 and self.entry_i==14:
                    return 0, False, False
                else:
                    return self.perms, self.recursive, False
            elif ch in (ord('i'), ord('I')): # and self.n>1
                return 0, False, False
            elif ch in (ord('a'), ord('A')): # and self.n>1
                return self.perms, self.recursive, True
            else:
                curses.beep()

    def run(self):
        selected = self.manage_keys()
        self.pwin.hide()
        curses.panel.update_panels()
        return selected


class DialogOwner:
    """Dialog to change file owner/group"""

    def __init__(self, filename, owner, group, owners, groups, i=0, n=0):
        self.h, self.w = 7+4, 45+4
        x0, y0 = int((app.w-self.w)/2), int((app.h-self.h)/2)
        try:
            self.win = curses.newwin(self.h, self.w, y0, x0)
            self.pwin = curses.panel.new_panel(self.win)
            self.pwin.top()
        except curses.error:
            raise
        self.win.keypad(1)
        self.win.bkgd(app.CLR['dialog'])
        self.file = filename
        self.owner = self.owner_old = owner
        self.group = self.group_old = group
        self.owners, self.groups = owners, groups
        self.recursive = True
        self.i, self.n, self.entry_i = i, n, 0

    def show_btns(self):
        attr_sel, attr_no = app.CLR['button_active'], app.CLR['button_inactive']
        self.win.addstr(self.h-2, self.w-21, '[<Ok>]', attr_sel if self.entry_i==11 else attr_no)
        self.win.addstr(self.h-2, self.w-13, '[ Cancel ]', attr_sel if self.entry_i==12 else attr_no)
        if self.n > 1:
            self.win.addstr(self.h-2, 3, '[ All ]', attr_sel if self.entry_i==13 else attr_no)
            self.win.addstr(self.h-2, 12, '[ Ignore ]', attr_sel if self.entry_i==14 else attr_no)

    def __fmt_text(self, text, n=10, ch='-'):
        l = len(text)
        return text[:n] if l>n else text+ch*(n-l)

    def show(self):
        self.win.erase()
        self.win.box()
        attr, attr_sel, attr_no = app.CLR['dialog_title'], app.CLR['dialog_perms'], app.CLR['dialog']
        title = 'Change file(s) owner/group'
        self.win.addstr(0, int((self.w-len(title)-2)/2), ' {} '.format(title), attr)
        if self.n > 1:
            buf = '{}/{}'.format(self.i, self.n)
            lbuf = len(buf)
            self.win.addstr(2, self.w-2-len(buf), buf)
        else:
            lbuf = 0
        self.win.addstr(2, 2, 'File:')
        self.win.addstr(2, 8, text2wrap(self.file, self.w-11-lbuf, fill=True), attr)
        self.win.addstr(4, 2, '          owner         group       recursive')
        self.win.addstr(5, 2, 'new:  [----------]  [----------]       [ ]')
        self.win.addstr(6, 2, 'old:  [----------]  [----------]       [ ]')
        self.win.addstr(5, 9, self.__fmt_text(self.owner), attr_sel if self.entry_i==0 else attr_no)
        self.win.addstr(5, 23, self.__fmt_text(self.group), attr_sel if self.entry_i==1 else attr_no)
        self.win.addstr(5, 42, 'X' if self.recursive else ' ', attr_sel if self.entry_i==3 else attr_no)
        self.win.addstr(6, 9, self.owner_old)
        self.win.addstr(6, 23, self.group_old)
        self.show_btns()
        self.win.refresh()

    def manage_keys(self):
        y, x = self.pwin.window().getbegyx()
        order = [0, 1, 3, 13, 14, 11, 12] if self.n>1 else [0, 1, 3, 11, 12]
        nels = len(order)
        while True:
            self.show()
            ch = self.win.getch()
            if ch in (0x03, 0x1B, ord('q'), ord('Q')):
                return -1, -1, False, False
            elif ch in (ord('\t'), 0x09, curses.KEY_DOWN, curses.KEY_RIGHT):
                i = order.index(self.entry_i)
                self.entry_i = order[0 if i==nels-1 else i+1]
            elif ch in (curses.KEY_UP, curses.KEY_LEFT, curses.KEY_BTAB):
                i = order.index(self.entry_i)
                self.entry_i = order[nels-1 if i==0 else i-1]
            elif ch in (ord(' '), 0x0A, 0x0D):
                if self.entry_i == 3:
                    self.recursive = not self.recursive
                elif self.entry_i == 0:
                    ret = SelectItem('Select new owner', self.owners, y+6, x+7, self.owner).run()
                    if ret != -1:
                        self.owner = ret
                elif self.entry_i == 1:
                    ret = SelectItem('Select new group', self.groups, y+6, x+21, self.group).run()
                    if ret != -1:
                        self.group = ret
                elif self.entry_i == 12:
                    return -1, -1, False, False
                elif self.n>1 and self.entry_i==13:
                    return self.owner, self.group, self.recursive, True
                elif self.n>1 and self.entry_i==14:
                    return 0, 0, False, False
                else:
                    return self.owner, self.group, self.recursive, False
            elif ch in (ord('i'), ord('I')): # and self.n>1
                return 0, 0, False, False
            elif ch in (ord('a'), ord('A')): # and self.n>1
                return self.owner, self.group, self.recursive, True
            else:
                curses.beep()

    def run(self):
        selected = self.manage_keys()
        self.pwin.hide()
        curses.panel.update_panels()
        return selected


########################################################################
##### TreeView
class TreeView:
    """TreeView class"""

    def __init__(self, path=sep):
        if not exists(path) or not isdir(path):
            raise ValueError('Path does not exist or is not dir: "{}"'.format(path))
        if path[-1] == sep and path != sep:
            path = path[:-1]
        self.path = path
        self.tree = DirsTree(path, app.cfg.options.show_dotfiles)
        self.__init_ui()

    def __init_ui(self):
        """initialize curses stuff"""
        self.__calculate_dims()
        try:
            self.win = curses.newwin(*self.dims)
        except curses.error:
            raise
        self.win.keypad(1)
        if curses.has_colors():
            self.win.bkgd(app.CLR['files_reg'])

    def __calculate_dims(self):
        if app.pane_active.mode == PaneMode.half:
            win = app.pane_inactive.win
        elif app.pane_active.mode == PaneMode.full:
            win = app.pane_active.win
        h, w = win.getmaxyx()
        y0, x0 = win.getbegyx()
        self.dims = h, w, y0, x0

    def display(self):
        h, n = app.h-4, len(self.tree)
        j, a, z = 0, self.tree.pos//h * h, (self.tree.pos//h+1) * h
        self.win.erase()
        self.win.attrset(app.CLR['pane_active'])
        self.win.box()
        display_scrollbar(self.win, 1, self.dims[1]-1, h, n, a, a)
        self.win.addstr(0, 2, ' Tree ', app.CLR['pane_header_path'])
        self.win.attrset(app.CLR['files_reg'])
        if z > n:
            a, z = max(n-h, 0), n
        for i in range(a, z):
            j += 1
            name, depth, fullname = self.tree[i]
            if name == sep:
                self.win.addstr(j, 1, ' ')
            else:
                self.win.move(j, 1)
                for kk in range(depth):
                    self.win.addstr(' ')
                    self.win.addch(curses.ACS_VLINE) # \u2502
                    self.win.addstr(' ')
                self.win.addstr(' ')
                if i == n-1:
                    self.win.addch(curses.ACS_LLCORNER) # \u2514
                elif depth > self.tree.get_depth(i+1):
                    self.win.addch(curses.ACS_LLCORNER) # \u2514
                else:
                    self.win.addch(curses.ACS_LTEE) # \u251C
                self.win.addch(curses.ACS_HLINE) # \u2500
                self.win.addstr(' ')
            w, wd = app.w//2-2, 3*depth+4
            if fullname == self.path:
                self.win.addstr(text2wrap(name, w-wd-3, fill=False), app.CLR['cursor'])
                if self.tree.has_children_dirs:
                    self.win.addstr(' \u27A9') # ' ->'
            else:
                self.win.addstr(text2wrap(name, w-wd, fill=False))
        app.statusbar.show_message('Path: {}'.format(self.path))

    def run(self):
        while True:
            self.display()
            chext = 0
            ch = self.win.getch()
            # to avoid extra chars input
            if ch == 0x1B:
                chext = 1
                ch = self.win.getch()
                ch = self.win.getch()
            if ch in (ord('k'), ord('K'), curses.KEY_UP):
                if self.tree.pos == 0 or self.tree.is_first_sibling:
                    continue
                newpos = self.tree.pos - 1
            elif ch in (ord('j'), ord('j'), curses.KEY_DOWN):
                if self.tree.pos == len(self.tree)-1 or self.tree.is_last_sibling:
                    continue
                newpos = self.tree.pos + 1
            elif ch in (curses.KEY_PPAGE, curses.KEY_BACKSPACE, 0x02):
                if self.tree.pos-(app.h-4) >= 0:
                    if self.tree.cur_depth == self.tree.get_depth(self.tree.pos-(app.h-4)):
                        newpos = self.tree.pos-(app.h-4)
                    else:
                        newpos = self.tree.first_sibling_pos
                else:
                    newpos = self.tree.first_sibling_pos
            elif ch in (curses.KEY_NPAGE, ord(' '), 0x06):   # Ctrl-F
                if self.tree.pos+(app.h-4) <= len(self.tree)-1:
                    if self.tree.cur_depth == self.tree.get_depth(self.tree.pos+(app.h-4)):
                        newpos = self.tree.pos+(app.h-4)
                    else:
                        newpos = self.tree.last_sibling_pos
                else:
                    newpos = self.tree.last_sibling_pos
            elif (ch in (curses.KEY_HOME, 0x01)) or (chext == 1) and (ch == 72):
                newpos = 1
            elif (ch in (curses.KEY_END, 0x05)) or (chext == 1) and (ch == 70):
                newpos = len(self.tree) - 1
            elif ch == curses.KEY_LEFT:
                if self.tree.pos == 0:
                    continue
                newpos = self.tree.parent_pos
            elif ch == curses.KEY_RIGHT:
                new_path = self.tree.to_child()
                if new_path is not None:
                    self.path = new_path
                continue
            elif ch in (10, 13):
                return self.path
            elif ch == curses.KEY_RESIZE:
                curses.doupdate()
                app.resize()
                self.__calculate_dims()
                self.win.resize(self.dims[0], self.dims[1])
                self.win.mvwin(self.dims[2], self.dims[3])
                continue
            elif ch in (ord('q'), ord('Q'), curses.KEY_F10, 0x03):  # Ctrl-C
                return -1
            else:
                continue
            # update
            self.path = self.tree.regenerate_from_pos(newpos)


########################################################################
##### InternalView
class InternalView:
    """View information on full screen"""

    def __init__(self, title, lbuf, center=True):
        self.title = title
        self.prepare_lines(lbuf, center)
        self.init_curses()

    def prepare_lines(self, lbuf, center):
        self.lbuf = [(text2wrap(l, app.w-2).strip(), c) for l, c in lbuf]
        self.nlines = len(lbuf)
        self.large = self.nlines > app.h-2
        if center:
            col_max = max_length([l for l, _ in self.lbuf])
            self.x0, self.y0 = (app.w-col_max)//2, 0 if self.large else (app.h-2-self.nlines)//2
        else:
            self.x0, self.y0 = 1, 0 if self.large else 1
        self.y = 0

    def init_curses(self):
        try:
            win_title = curses.newwin(1, app.w, 0, 0)
            self.win_body = curses.newwin(app.h-2, app.w, 1, 0)
            win_status = curses.newwin(1, app.w, app.h-1, 0)
        except curses.error:
            raise
        win_title.bkgd(app.CLR['header'])
        self.win_body.bkgd(app.CLR['view_white_on_black'])
        win_status.bkgd(app.CLR['statusbar'])
        self.win_body.keypad(1)
        win_title.erase()
        win_status.erase()
        title = text2wrap(self.title, app.w-1).strip()
        win_title.addstr(0, (app.w-length(title))//2, title)
        status = '' if self.large else 'Press any key to continue'
        win_status.addstr(0, (app.w-len(status))//2, status)
        win_title.refresh()
        win_status.refresh()

    def show(self):
        self.win_body.erase()
        for i, (l, c) in enumerate(self.lbuf[self.y:self.y+app.h-2]):
            self.win_body.addstr(self.y0+i, self.x0, l, app.CLR[c])
        self.win_body.refresh()

    def run(self):
        if self.large:
            while True:
                self.show()
                ch = self.win_body.getch()
                if ch in (ord('k'), ord('K'), curses.KEY_UP):
                    self.y = max(self.y-1, 0)
                if ch in (ord('j'), ord('J'), curses.KEY_DOWN):
                    self. y = min(self.y+1, self.nlines-1)
                elif ch in (curses.KEY_HOME, 0x01):
                    self.y = 0
                elif ch in (curses.KEY_END, 0x05):
                    self.y = self.nlines - 1
                elif ch in (curses.KEY_PPAGE, 0x08, 0x02, curses.KEY_BACKSPACE):
                    self.y = max(self.y-app.h+2, 0)
                elif ch in (curses.KEY_NPAGE, ord(' '), 0x06):
                    self.y = min(self.y+app.h-2, self.nlines-1)
                elif ch in (0x1B, ord('q'), ord('Q'), curses.KEY_F3, curses.KEY_F10):
                    break
        else:
            self.show()
            while not self.win_body.getch():
                pass


########################################################################



# -*- coding: utf-8 -*-


import errno
import curses
from os import getuid, pardir
from os.path import dirname, exists, join
from datetime import datetime

from preferences import Config, load_colortheme, load_keys, History
from folders import new_folder, is_delete_oldfs
from utils import num2str, length, text2wrap, get_realpath, run_in_background, run_shell, ProcessCommand
from ui_widgets import display_scrollbar, DialogError, DialogConfirm, EntryLine, InternalView
from key_defs import key_bin2str
from actions import do
from common import *


########################################################################
##### Definitions and module variables
colors_table = {'black': curses.COLOR_BLACK,
                'blue': curses.COLOR_BLUE,
                'cyan': curses.COLOR_CYAN,
                'green': curses.COLOR_GREEN,
                'magenta': curses.COLOR_MAGENTA,
                'red': curses.COLOR_RED,
                'white': curses.COLOR_WHITE,
                'yellow': curses.COLOR_YELLOW}

##### Module variables
app, cfg = None, None


########################################################################
##### Main window
class UI:
    def __init__(self, cfg, win, paths1, paths2):
        logging.debug('Create UI')
        self.cfg = cfg
        self.win = win
        self.w = self.h = 0
        self.CLR = dict()
        self.init_curses()
        self.pane1 = Pane(self, paths1)
        self.pane2 = Pane(self, paths2)
        self.statusbar = StatusBar(self)
        self.cli = PowerCLI(self)
        self.focus_pane(self.pane1)
        self.resize()

    def init_curses(self):
        curses.cbreak()
        curses.raw()
        self.win.leaveok(1)
        self.win.keypad(1)
        curses.curs_set(0)
        self.init_colors()
        self.init_keys()
        self.init_history()
        # HACK: ugly hack to inject main app in that module
        import ui_widgets
        ui_widgets.app = self

    def init_colors(self):
        if curses.has_colors():
            try:
                colors = load_colortheme()
            except FileNotFoundError:
                raise
            for i, col in enumerate(colors.keys()):
                fg, bg = colors[col]
                light = fg.endswith('*')
                fg = fg[:-1] if fg.endswith('*') else fg
                bg = bg[:-1] if bg.endswith('*') else bg
                color_fg, color_bg = colors_table[fg], colors_table[bg]
                curses.init_pair(i+1, color_fg, color_bg)
                self.CLR[col] = curses.color_pair(i+1)
                if light:
                    self.CLR[col] = self.CLR[col] | curses.A_BOLD
        else:
            for col in COLOR_ITEMS:
                self.CLR[col] = curses.color_pair(0)

    def init_keys(self):
        try:
            self.keys = load_keys()
        except FileNotFoundError:
            raise

    def init_history(self):
        self.history = History()
        if self.cfg.options.save_history_at_exit:
            try:
                self.history.load()
            except Exception as e:
                return

    def focus_pane(self, pane):
        pane.focus = True
        otherpane = self.pane2 if pane==self.pane1 else self.pane1
        otherpane.focus = False

    @property
    def pane_active(self):
        return self.pane1 if self.pane1.focus else self.pane2

    @property
    def pane_inactive(self):
        return self.pane2 if self.pane1.focus else self.pane1

    def resize(self):
        h, w = self.win.getmaxyx()
        logging.debug('Resize UI: w={}, h={}'.format(w, h))
        self.h, self.w = h, w
        if w == 0 or h == 2:
            return
        if w < MIN_COLUMNS:
            raise LFMTerminalTooNarrow
        self.win.resize(self.h, self.w)
        if self.pane1.mode == PaneMode.full:
            self.pane1.resize(0, 0, h-1, w)
        elif self.pane1.mode == PaneMode.hidden:
            self.pane1.resize(0, 0, h-1, w)
        else:
            self.pane1.resize(0, 0, h-1, w//2)
        if self.pane2.mode == PaneMode.full:
            self.pane2.resize(0, 0, h-1, w)
        elif self.pane2.mode == PaneMode.hidden:
            self.pane2.resize(0, 0, h-1, w)
        else:
            self.pane2.resize(0, w//2, h-1, w//2)
        self.statusbar.resize(h-1, w)
        self.cli.resize(h-1, w)
        self.display()

    def clear_screen(self):
        logging.debug('Clear screen')
        self.pane1.clear()
        self.pane2.clear()
        self.statusbar.clear()
        curses.doupdate()

    def display(self):
        logging.debug('Display UI')
        self.pane1.display()
        self.pane2.display()
        self.display_statusbar_or_powercli()
        self.win.noutrefresh()
        curses.doupdate()

    def display_half(self):
        logging.debug('Display half UI')
        self.pane_active.display()
        self.display_statusbar_or_powercli()
        self.win.noutrefresh()
        curses.doupdate()

    def display_statusbar_or_powercli(self):
        if self.cli.visible:
            self.cli.display()
        else:
            self.statusbar.display()

    def run(self):
        self.display()
        while True:
            ret, extra = self.get_key()
            if ret == RetCode.quit_chdir:
                return extra
            elif ret == RetCode.quit_nochdir:
                return None
            elif ret == RetCode.fix_limits:
                self.pane_active.tab_active.fix_limits()
                self.display_half()
            elif ret == RetCode.full_redisplay:
                self.display()
            elif ret == RetCode.half_redisplay:
                self.display_half()

    def get_key(self):
        self.win.nodelay(False)
        km = KeyModifier.none
        key = self.win.getch()
        if key == 27: # Esc or Alt
            km = KeyModifier.alt
            self.win.nodelay(True)
            key = self.win.getch()
            if key == -1: # Esc
                km, key = KeyModifier.none, 27
        if key == curses.KEY_RESIZE:
            self.resize()
            return RetCode.full_redisplay, None
        key_str = key_bin2str((km, key))
        action = self.keys[(km, key)] if (km, key) in self.keys else None
        logging.debug('Key pressed: {0} [{1:#x}] => {2} => {3} -> {4}'
                      .format(curses.keyname(key), key, str((km, key)), key_str, action if action else '<undefined>'))
        return do(self, action) if action else (RetCode.nothing, None)


########################################################################
##### Pane
class Pane:
    def __init__(self, ui, paths):
        logging.debug('Create Pane')
        self.ui = ui
        self.mode = PaneMode.half
        self.focus = False
        self.tabs = [Tab(p) for p in paths[:MAX_TABS]]
        self.tab_active = self.tabs[0]
        # ui
        try:
            self.win_tabs = curses.newwin(1, 1, 0, 0)
            self.win = curses.newwin(10, 10, 0, 0)
        except curses.error:
            raise
        self.win.bkgd(self.ui.CLR['pane_inactive'])

    def resize(self, y0, x0, h, w):
        logging.debug('Resize Pane: x0={}, y0={}, w={}, h={}'.format(x0, y0, h, w))
        self.x0, self.y0 = x0, y0
        self.w, self.h = w, h
        self.fh = h-4 if self.mode==PaneMode.half else h-1
        self.win_tabs.resize(2, w)
        self.win_tabs.mvwin(y0, x0)
        self.win.resize(h-1, w)
        self.win.mvwin(y0+1, x0)
        try:
            for tab in self.tabs:
                tab.fix_limits()
        except AttributeError:
            pass

    def clear(self):
        self.win.erase()
        self.win_tabs.erase()
        self.win.noutrefresh()
        self.win_tabs.noutrefresh()

    def display(self):
        logging.debug('Display Pane')
        if self.mode == PaneMode.hidden:
            return
        tab = self.tab_active
        CLR = self.ui.CLR
        # tabs
        self.win_tabs.erase()
        self.win_tabs.addstr(0, 0, ' '*self.w, CLR['header'])
        wtab = self.w // MAX_TABS
        for i, t in enumerate(self.tabs):
            pathname = '/' if t.fs.basename=='' else t.fs.basename
            buf = '[' + text2wrap(pathname, wtab-2, start_pct=.5) + ']'
            self.win_tabs.addstr(0, wtab*i, buf, CLR['tab_active' if t==tab else 'tab_inactive'])
        # contens:
        self.win.erase()
        if self.mode == PaneMode.half:
            self.__display_panehalf(tab, CLR)
        elif self.mode == PaneMode.full:
            self.__display_panefull(tab, CLR)
        # refresh
        self.win_tabs.noutrefresh()
        self.win.noutrefresh()

    def __display_panehalf(self, tab, CLR):
        if self.focus:
            attr, attr_path = CLR['pane_active'], CLR['pane_header_path']
        else:
            attr, attr_path = CLR['pane_inactive'], CLR['pane_inactive']
        self.win.attrset(attr)
        # box
        self.win.box()
        self.win.addstr(0, 2, text2wrap(tab.fs.path_str, self.w-5, start_pct=.33, fill=False), attr_path)
        col2 = self.w - 14 # sep between size and date: w - len(mtime) - 2x borders
        col1 = col2 - 8    # sep between filename and size: col2 - len(size) - 1x border
        self.win.addstr(1, 1, 'Name'.center(col1-2)[:col1-2], CLR['pane_header_titles'])
        self.win.addstr(1, col1+2, 'Size', CLR['pane_header_titles'])
        self.win.addstr(1, col2+5, 'Date', CLR['pane_header_titles'])
        if tab.fs.cfg.show_dotfiles:
            self.win.addstr(0, self.w-3, '·', attr)
        if tab.fs.cfg.filters:
            self.win.addstr(0, self.w-2, 'f', attr)
        # files
        fmt = [('type', 1), ('name', col1-2), ('sep', 1), ('size', 7), ('sep', 1), ('mtime', 12)]
        for i in range(self.h-4):
            if i+tab.a >= len(tab.fs):
                break
            f = tab.fs[tab.a+i]
            if self.focus and tab.i == tab.a+i:
                attr = 'cursor_selected' if f in tab.selected else 'cursor'
            else:
                attr = 'selected_files' if f in tab.selected else 'files_' + f.get_type_from_ext(self.ui.cfg.files_ext)
            self.win.addstr(2+i, 1, f.format(fmt), CLR[attr])
        # vertical separators
        self.win.vline(1, col1, curses.ACS_VLINE, self.h-3)
        self.win.vline(1, col2, curses.ACS_VLINE, self.h-3)
        if self.focus:
            self.win.vline(tab.i-tab.a+2, col1, curses.ACS_VLINE, 1, CLR['cursor'])
            self.win.vline(tab.i-tab.a+2, col2, curses.ACS_VLINE, 1, CLR['cursor'])
        # scrollbar
        display_scrollbar(self.win, 2, self.w-1, self.h-4, len(tab.fs), tab.i, tab.a)

    def __display_panefull(self, tab, CLR):
        self.win.attrset(CLR['pane_inactive'])
        # files
        col = self.w - 64 # 1x border
        fmt = [('type', 1), ('mode', 9), ('sep', 2), ('owner', 10), ('sep', 2), ('group', 10), ('sep', 2),
               ('size', 7), ('sep', 2), ('mtime2', 16), ('sep', 2), ('name', col)]
        for i in range(self.h-1):
            if i+tab.a >= len(tab.fs):
                break
            f = tab.fs[tab.a+i]
            if tab.i == tab.a+i:
                attr = 'cursor_selected' if f in tab.selected else 'cursor'
            else:
                attr = 'selected_files' if f in tab.selected else 'files_' + f.get_type_from_ext(self.ui.cfg.files_ext)
            self.win.addstr(i, 0, f.format(fmt, sep=' '), CLR[attr])
        # scrollbar
        display_scrollbar(self.win, 0, self.w-1, self.h-1, len(tab.fs), tab.i, tab.a)

    def change_mode(self, newmode):
        otherpane = self.ui.pane2 if self==self.ui.pane1 else self.ui.pane1
        if self.mode == PaneMode.full:
            otherpane.mode = PaneMode.half
        if newmode == PaneMode.full:
            otherpane.mode = PaneMode.hidden
        self.mode = newmode
        self.ui.resize()

    def refresh(self):
        for tab in self.tabs:
            tab.refresh()

    def insert_new_tab(self, path, lefttab):
        newtab = Tab(path)
        self.tabs.insert(self.tabs.index(lefttab)+1, newtab)
        self.tab_active = newtab

    def close_tab(self, tab):
        idx = self.tabs.index(tab)
        tab.close()
        self.tabs.remove(tab)
        self.tab_active = self.tabs[idx-1]


########################################################################
##### Tab
class Tab:
    def __init__(self, path):
        self.fs = None
        self.history = []
        self.goto_folder(path)

    def __check_rebuild(self):
        if self.fs.vfs:
            rebuild = 1 if app.cfg.options.rebuild_vfs is True else 0
            if app.cfg.confirmations.ask_rebuild_vfs:
                rebuild = DialogConfirm('Rebuild vfs file', self.fs.base_filename, rebuild)
            return rebuild==1
        else:
            return False

    def close(self):
        self.fs.exit(all_levels=True, rebuild=self.__check_rebuild())

    def goto_folder(self, path, delete_vfs_tree=False, files=None):
        """Called when chdir to a new path"""
        oldpath = self.fs.path_str if self.fs else None
        oldfs = self.fs
        try:
            if files: # search vfs
                self.fs = new_folder(path, self.fs, files=files)
            else:
                rebuild = self.__check_rebuild() if is_delete_oldfs(path, self.fs) else False
                self.fs = new_folder(path, self.fs, rebuild_if_exit=rebuild)
            self.fs.cfg.filters = oldfs.cfg.filters if oldfs else ''
        except PermissionError as e:
            logging.warning('ERROR: cannot enter in {}: {}'.format(path, str(e)))
            app.display()
            DialogError('Cannot chdir {}\n{}'.format(path, str(e).split(':', 1)[0]))
            return
        except UserWarning as e:
            logging.warning('ERROR: cannot enter in {}: {}'.format(path, str(e)))
            app.display()
            DialogError('Cannot chdir {}\n{}'.format(path, str(e)))
            return
        except FileNotFoundError:
            logging.warning('ERROR: cannot enter in {}: invalid directory?'.format(path))
            app.display()
            DialogError('Cannot chdir {}\nInvalid directory?'.format(path))
            return
        if delete_vfs_tree:
            oldfs.exit(all_levels=True)
        self.a = self.i = 0
        self.selected = []
        self.refresh(first=True)
        if oldpath and VFS_STRING not in oldpath:
            if oldpath in self.history:
                self.history.remove(oldpath)
            self.history.append(oldpath)
            self.history = self.history[-HISTORY_MAX:]

    def reload(self):
        """Called when contents have changed"""
        try:
            self.fs.load()
        except OSError as err:
            if err.errno == errno.ENOENT: # dir deleted?
                pardir = self.fs.pdir
                while not exists(pardir):
                    pardir = dirname(pardir)
                self.goto_folder(pardir)
        self.refresh()

    def refresh(self, first=False):
        """Called when config or filters have changed"""
        if not first:
            oldi = self.i
            oldf = self.fs[self.i].name
            oldselected = [f.name for f in self.selected]
        self.fs.cfg.fill_with_app(cfg)
        self.fs.refresh()
        self.n = len(self.fs)
        i = 0 if first else self.fs.pos(oldf)
        self.i = oldi if i==-1 else i
        self.a = divmod(self.i, self.n)[0] * self.n
        if not first:
            self.fix_limits()
            self.selected = list(filter(None, [self.fs.lookup(f) for f in oldselected]))

    def fix_limits(self):
        self.i = max(0, min(self.i, self.n-1))
        self.a = int(self.i//app.pane_active.fh * app.pane_active.fh)

    def focus_file(self, filename):
        i = self.fs.pos(filename)
        if i != -1:
            self.i = i
            self.fix_limits()

    @property
    def dirname(self):
        return self.fs.pdir

    @property
    def current_filename(self):
        return self.fs[self.i].name

    @property
    def current_filename_full(self):
        return join(self.fs.pdir, self.fs[self.i].name)

    @property
    def current(self):
        return self.fs[self.i]

    @property
    def selected_or_current(self):
        if len(self.selected) == 0:
            cur = self.fs[self.i]
            return list() if cur.name==pardir else [cur]
        else:
            return self.selected

    @property
    def selected_or_current2(self):
        return self.selected if self.selected else [self.fs[self.i]]


########################################################################
##### Statusbar
class StatusBar:
    def __init__(self, ui):
        logging.debug('Create StatusBar')
        self.ui = ui
        try:
            self.win = curses.newwin(1, 10, 1, 0)
        except curses.error:
            raise
        self.win.bkgd(self.ui.CLR['statusbar'])

    def resize(self, y0, w):
        logging.debug('Resize StatusBar: y0={}, w={}'.format(y0, w))
        self.y0, self.w = y0, w
        self.win.resize(1, w)
        self.win.mvwin(y0, 0)

    def clear(self):
        self.win.erase()
        self.win.noutrefresh()

    def display(self):
        logging.debug('Display StatusBar')
        self.win.erase()
        tab = self.ui.pane_active.tab_active
        if len(tab.selected) > 0:
            if self.w >= 45:
                size = sum([f.size for f in tab.selected])
                self.win.addstr('    %s bytes in %d files' % (num2str(size), len(tab.selected)))
        else:
            if self.w >= 80:
                s = 'File: %4d of %-4d' % (tab.i+1, tab.n)
                if tab.fs.cfg.filters:
                    s+= '  [%d filtered]' % tab.fs.nfiltered
                self.win.addstr(s)
                filename = text2wrap(get_realpath(tab), self.w-20-len(s), fill=False)
                self.win.addstr(0, len(s)+4, 'Path: ' + filename)
        if self.w > 10:
            try:
                self.win.addstr(0, self.w-8, 'F1=Help')
            except:
                pass
        self.win.refresh()

    def show_message(self, text):
        self.win.erase()
        self.win.addstr(0, 1, text2wrap(text, self.w-2, fill=False))
        self.win.refresh()


########################################################################
##### PowerCLI
class PowerCLI:
    RUN_NORMAL, RUN_BACKGROUND, RUN_NEEDCURSESWIN = range(3)

    def __init__(self, ui):
        logging.debug('Create PowerCLI')
        self.ui = ui
        self.visible = self.running = False
        try:
            self.win = curses.newwin(1, 10, 1, 0)
        except curses.error:
            raise
        self.win.bkgd(self.ui.CLR['powercli_text'])
        self.entry, self.text, self.pos = None, '', 0

    def toggle(self):
        if self.visible:
            self.visible = self.running = False
        else:
            self.visible, self.running = True, False

    def resize(self, y0, w):
        logging.debug('Resize PowerCLI: y0={}, w={}'.format(y0, w))
        self.y0, self.w = y0, w
        self.win.resize(1, w)
        self.win.mvwin(y0, 0)

    def display(self):
        if self.running:
            return
        logging.debug('Display PowerCLI')
        tab = self.ui.pane_active.tab_active
        self.win.erase()
        path = text2wrap(tab.fs.path_str, self.ui.w//6, start_pct=0, fill=False)
        prompt = '[{}]{} '.format(path, '#' if getuid()==0 else '$')
        lprompt = length(prompt)
        self.win.addstr(0, 0, prompt, self.ui.CLR['powercli_prompt'])
        self.win.noutrefresh()
        curses.curs_set(1)
        self.running = True
        self.entry = EntryLine(self, self.w-lprompt, self.y0, lprompt, self.text,
                               history=self.ui.history['cli'][:], is_files=True, cli=True)
        self.entry.pos = self.pos
        self.entry.show()
        ans = self.entry.manage_keys()
        if ans == -1:           # Ctrl-C
            cmd, self.text, self.pos = None, '', 0
        elif ans == -2:           # Ctrl-X
            cmd, self.text, self.pos = None, self.entry.text, self.entry.pos
        elif ans == 10:         # return
            cmd, self.text, self.pos = self.entry.text.strip(), '', 0
            if cmd:
                self.execute(cmd)
        else:
            raise ValueError
        curses.curs_set(0)
        self.visible = False
        self.ui.display()

    def execute(self, cmd):
        self.ui.history.append('cli', cmd)
        logging.debug('PowerCLI Execute: |{}|'.format(cmd))
        selected = [f for f in self.ui.pane_active.tab_active.selected_or_current2]
        if cmd[-1] == '&':
            mode, cmd = PowerCLI.RUN_BACKGROUND, cmd[:-1].strip()
        elif cmd[-1] == '$':
            mode, cmd = PowerCLI.RUN_NEEDCURSESWIN, cmd[:-1].strip()
        else:
            mode = PowerCLI.RUN_NORMAL
        for f in selected:
            try:
                cmd2 = self.__replace_cli(cmd, f)
            except Exception as err:
                log.warning('Cannot execute PowerCLI command: {}\n{}'.format(cmd2, str(err)))
                DialogError('Cannot execute PowerCLI command:\n  {}\n\n{}'.format(cmd2, str(err)))
            else:
                if self.__run(cmd2, self.ui.pane_active.tab_active.dirname, mode) == -1:
                    self.ui.display()
                    if DialogConfirm('Error running PowerCLI', 'Do you want to stop now?') == 1:
                        break
        self.ui.pane_active.tab_active.selected = []

    def __replace_cli(self, cmd, f):
        # prepare variables
        tab = self.ui.pane_active.tab_active
        filename = f.name
        cur_directory = tab.dirname
        other_directory = self.ui.pane_inactive.tab_active.dirname
        fullpath = f.pfile
        filename_noext = f.name_noext
        ext = f.ext
        all_selected = [s.name for s in tab.selected]
        all_files = [elm for elm in tab.fs.get_filenames() if elm is not pardir]
        try:
            selection_idx = all_selected.index(filename)+1
        except ValueError:
            selection_idx = 0
        tm = datetime.fromtimestamp(f.mtime)
        ta = datetime.fromtimestamp(f.stat.st_atime)
        tc = datetime.fromtimestamp(f.stat.st_ctime)
        tnow = datetime.now()
        dm = tm.date()
        da = ta.date()
        dc = tc.date()
        dnow = tnow.date()
        # conversion table
        lcls = {'f': filename, 'v': filename, 'F': fullpath,
                'E': filename_noext, 'e': ext,
                'p': cur_directory, 'o': other_directory,
                's': all_selected, 'a': all_files, 'i': selection_idx,
                'dm': dm, 'da': da, 'dc': dc, 'dn': dnow,
                'tm': tm, 'ta': ta, 'tc': tc, 'tn': tnow}
        for k, bmk in self.ui.cfg.bookmarks.items():
            lcls['b{}'.format(k)] = bmk
        # and replace, first python code, and then variables
        cmd = self.__replace_python(cmd, lcls)
        cmd = self.__replace_variables(cmd, lcls)
        return cmd

    def __replace_python(self, cmd, lcls):
        lcls = dict([('__lfm_{}'.format(k), v) for k, v in lcls.items()])
        # get chunks
        chunks, st = {}, 0
        while True:
            i = cmd.find('{', st)
            if i == -1:
                break
            j = cmd.find('}', i+1)
            if j == -1:
                raise SyntaxError('{ at %d position has not ending }' % i)
            else:
                chunks[(i+1, j)] = cmd[i+1:j].replace('$', '__lfm_')
                st = j + 1
        # evaluate
        if chunks == {}:
            return cmd
        buf, st = '', 0
        for i, j in sorted(chunks.keys()):
            buf += cmd[st:i-1]
            try:
                translated = eval(chunks[(i, j)], {}, lcls)
            except Exception as err:
                raise SyntaxError(str(err).replace('__lfm_', '$'))
            buf += translated
            st = j+1
        buf += cmd[st:]
        return buf

    def __replace_variables(self, cmd, lcls):
        for k, v in lcls.items():
            if k in ('i', ):
                cmd = cmd.replace('${}'.format(k), str(v))
            elif k in ('dm', 'da', 'dc', 'dn', 'tm', 'ta', 'tc', 'tn'):
                cmd = cmd.replace('${}'.format(k), str(v).split('.')[0])
            elif k in ('s', 'a'):
                cmd = cmd.replace('${}'.format(k), ' '.join(['"{}"'.format(f) for f in v]))
            else:
                cmd = cmd.replace('${}'.format(k), v)
        return cmd

    def __run(self, cmd, path, mode):
        curses.curs_set(0)
        if mode == PowerCLI.RUN_NEEDCURSESWIN:
            run_shell(cmd, path)
            st, msg, err = 0, '', ''
        elif mode == PowerCLI.RUN_BACKGROUND:
            run_in_background(cmd, path)
            st, msg, err = 0, '', ''
        else: # PowerCLI.RUN_NORMAL
            st, msg, err = ProcessCommand('Executing PowerCLI', cmd, cmd, path).run()
        if err:
            log.warning('Error running PowerCLI command: {}\n{}'.format(cmd, str(err)))
            DialogError('Error running PowerCLI command:\n  {}\n\n{}'.format(cmd, str(err)))
        if st != -100 and msg:
            if self.ui.cfg.options.show_output_after_exec:
                if DialogConfirm('Executing PowerCLI', 'Show output?', 1) == 1:
                    lst = [(l, 'view_white_on_black') for l in msg.split('\n')]
                    InternalView('Output of: {}'.format(cmd), lst, center=False).run()
        return st

########################################################################
##### Main
def init_config(use_wide_chars):
    global cfg
    cfg = Config()
    cfg.load()
    # HACK: ugly hack to inject option in that module
    import utils
    utils.use_wide_chars = True if use_wide_chars else cfg.options.use_wide_chars
    log.info('Support wide chars: {}'.format(utils.use_wide_chars))
    return cfg


def launch_ui(win, path1, path2, use_wide_chars):
    global app
    cfg = init_config(use_wide_chars)
    try:
        app = UI(cfg, win, [path1], [path2])
        ret = app.run()
    except LFMTerminalTooNarrow as e:
        DialogError('Terminal too narrow to show contents.\nIt should have {} columns at mininum.'.format(MIN_COLUMNS))
        return None
    if app.cfg.options.save_configuration_at_exit:
        try:
            app.cfg.save()
        except Exception as e:
            DialogError('Cannot save configuration file\n{}'.format(str(e)))
    if app.cfg.options.save_history_at_exit:
        try:
            app.history.save()
        except Exception as e:
            DialogError('Cannot save history file\n{}'.format(str(e)))
    return ret


def run_app(path1, path2, use_wide_chars):
    return curses.wrapper(launch_ui, path1, path2, use_wide_chars)


########################################################################
